from sys import stdout

# *******************************************************************************
# 1st-order approximation ageing model
def ageing(alpha, tau):
	K = 0.3 # fitting parameter by assuming 10 yrs degradation equalling to 20%
	t = 10 # degradation for 10yrs
	delta_tau = K * tau * (alpha * t)**(1.0/6)
	return delta_tau

# *******************************************************************************


# *******************************************************************************
# work out cell/path delay degradation
def apply_ageing(alpha, tau, transition, delay_acc):

	delta_tau = []		# Degradation of cell delay
	delta_total = []	# Degradation of path delay
	delta_percent = []	# Percentage degradation of path delay
	path_delay = []		# Path delays

	# for each path 
	for path_index, path in enumerate(tau):
		delta_tau.append([])
		
		# for each cell of the path
		for cell_index, cell in enumerate(path):
			
			# if output rising (PMOS stressed)
			if transition[path_index][cell_index+1] is 'r':
				
				delta = ageing(float(alpha[path_index][cell_index]), \
						float(tau[path_index][cell_index]))
			else:
				#delta = 0
				delta = 0.1 * ageing(1-float(alpha[path_index][cell_index]), \
						float(tau[path_index][cell_index]))
		
			delta_tau[path_index].append(delta)

		delta_total.append(sum(delta_tau[path_index]))
		
		D_acc = sum([float(i) for i in delay_acc[path_index]])
		delta_percent.append(100 * delta_total[path_index]/D_acc)

		path_delay.append(D_acc)

	return delta_tau, delta_total, delta_percent, path_delay

# *******************************************************************************


# *******************************************************************************
# find signal probabilities for all gate ports in path
def find_sp(all_node, all_sp, wire_name, iteration):
	
	sp = []
		
	for path_index, line in enumerate(wire_name):
		sp.append([])
		
		for gate in line:
		
			# add sp for all gate ports in path
			if isinstance(all_sp[0], list):
				sp[path_index].append(all_sp[iteration][all_node.index(gate)])
			else:
				sp[path_index].append(all_sp[all_node.index(gate)])
	
	return sp
		
# *******************************************************************************


# *******************************************************************************
# convert port name to wire name
def port2wire(port_name, instance_r, cell_in_r, wire_in_r):
	
	wire_name = []

	# for each path
	for i, path in enumerate(port_name):

		wire_name.append([])

		# for each port
		for port in path:
			instance = port.split('/')[0]
			inport = port.split('/')[1]

			# find instance
			instance_index = instance_r.index(instance)
		
			# find port
			port_index = cell_in_r[instance_index].index(inport)
		
			wire = wire_in_r[instance_index][port_index]	
		
			wire_name[i].append(wire)

	return wire_name

# *******************************************************************************


# ******************************************************************************
# statistics
def mean(data):
    """Return the sample arithmetic mean of data."""
    n = len(data)
    if n < 1:
        raise ValueError('mean requires at least one data point')
    return sum(data)/n # in Python 2 use sum(data)/float(n)

def _ss(data):
    """Return sum of square deviations of sequence data."""
    c = mean(data)
    ss = sum((x-c)**2 for x in data)
    return ss

def pstdev(data):
    """Calculates the population standard deviation."""
    n = len(data)
    if n < 2:
        raise ValueError('variance requires at least two data points')
    ss = _ss(data)
    pvar = ss/n # the population variance
    return pvar**0.5	
# ******************************************************************************


# *******************************************************************************
# iteratively applying ageing model
def iteration(number, nbti_path, port_in, cell_delay, cell_transition, in2in_delay):
	
	# path degradation
	delta = []

	# for each input vector
	for x in range(0, number):

		stdout.write('\rApplying ageing model to input vector %d/%d ...' %((x+1), number))
		stdout.flush()

		# find alpha for all gates
		path_alpha = []
		for path in nbti_path:
			alpha_all_cell = [alpha_all[x] for alpha_all in path]
			path_alpha.append(alpha_all_cell)

		# work out stress duty cycle for all gates
		delta_tau, delta_total, delta_percent, path_delay = apply_ageing(path_alpha, \
									cell_delay, \
									cell_transition, \
									in2in_delay)

		delta.append(delta_total)
	
	# degradation mean value
	delta_mean = [sum(x)/number for x in zip(*delta)]
	#delta_mean = [mean(x) for x in zip(*delta)]

	# degradation sigma
	#delta_std = [pstdev(x) for x in zip(*delta)]
	
	# degradation worst
	#delta_worst = [x+3*y for x,y in zip(delta_mean, delta_std)]
	#delta_worst = [max(x) for x in zip(*delta)]
	#delta_best = [x-3*y for x,y in zip(delta_mean, delta_std)]

	# lifetime mean value
	#lifetime_mean = [(delay_constraint-a)/b for a,b in zip(path_delay, delta_mean)]

	# minimal lifetime path index
	#min_lifetime_index = lifetime_mean.index(min(lifetime_mean))

	return delta_mean, path_delay, delta
# ******************************************************************************
