import ageing, general, lib_parser, math_parser, netlist_parser, sdf_parser, signal_probability, timing_parser, re, os, shutil, optimization, area_parser, delay_calculation, time, sys, random

# *******************************************************************************
# ageing-aware synthesis main program
# definition:
# ----------------------------------------------------------------------------
# outdir - output directory name
# netlist_file - test netlist
# netlist_rename - renamed netlist
# netlist_backup - backup netlist
# tcl - Design Compiler tcl
# timing_rpt - DC timing report
# area_rpt - DC area report
# sdf_file - DC sdf
# lib_file - technology library file
# vec_no - number of input vectors
# seed - seed for random input vectors generation
# sp_file - signal probability expression file
# log_path - optimization log
# max_area - area constraint

# preparing
outdir = '../dc_syn'
netlist_file = '../lib/st65/EPFL/log2_lo.v'
netlist_rename = '/test.v'
netlist_backup = '/design_opt.v'
tcl = '../scripts/upsizing65.tcl'
vec_no = 1
seed = 123
sp_file = '../lib/st65/st65.sp'
lib_file = '../lib/st65/CORE65LPSVT_nom_1.10V_25C.lib'
log_path = '/optimization.log'
stacking_file = '../lib/st65/st65_stacking.txt'

# eva
rise_timing_rpt = '/test_timing_rise.rpt'
fall_timing_rpt = '/test_timing_fall.rpt'
area_rpt = '/test_area.rpt'
sdf_file = '/test.sdf'
delay_constraint = 45.87

# trial
delay_log_path = '/delay.log'
delta_log_path = '/delta.log'
max_trial = 10
trial_seed = 5

#decision
sensitivity_type = 'degradation/area'
area_constraint = float('inf')
seed_2 = 1

netlist_r_text = '../lib/st65/EPFL/log2_lo_reorder.txt'


# *******************************************************************************
def prepare():		

	print 'start lifetime optimization - upsizing'

	# check if sensitivity type correct
	if sensitivity_type != 'delay/area' and sensitivity_type != 'degradation/area':
		print 'Error: unrecognized sensitivity type'
		sys.exit(0)

	# create output directory
	general.create_outdir(outdir)

	# move test netlist into output directory and rename
	shutil.copy(netlist_file, outdir+netlist_rename)

	# make a netlist backup
	shutil.copy(netlist_file, outdir+netlist_backup)
	
	print 'parsing library file ...'
	
	# read library file
	lib_content = general.read_txt(lib_file)

	# parse library file
	global lib
	lib = lib_parser.analyze(lib_content)
	
	print 'parsing netlist ...'

	# read netlist
	netlist_content = general.read_txt(outdir+netlist_rename)

	# remove '\.../' from netlist
	netlist_parser.virgule_removal(netlist_content, outdir+netlist_rename)

	# read netlist
	netlist_content = general.read_txt(outdir+netlist_rename)

	# remove '\.../' from netlist
	netlist_parser.virgule_removal(netlist_content, outdir+netlist_backup)

	# parse netlist
	netlist = netlist_parser.analyze(netlist_content, lib)

	print 'parse netlist complete'

	# design primary inputs
	primary_inputs = netlist.inputs

	# reorder netlist according connections
	global instance_r
	global wire_in_r
	global cell_in_r
	#instance_r, cell_typ, wire_in_r, wire_out, cell_in_r, cell_out = netlist_parser.reorder(netlist.instance, netlist.cell_typ, netlist.wire_in, netlist.wire_out, netlist.cell_in, netlist.cell_out, netlist.inputs)

	# store and restore reordered netlist
	#netlist_parser.reorder_store(instance_r, cell_typ, wire_in_r, wire_out, cell_in_r, cell_out, outdir)
	instance_r, cell_typ, wire_in_r, wire_out, cell_in_r, cell_out = netlist_parser.reorder_restore(netlist_r_text)

	# generate input vectors
	input_vec = signal_probability.generate(len(primary_inputs), vec_no, seed)
	
	# read sp expression file
	sp_content = general.read_txt(sp_file)

	# parse sp expression
	sp_typ, sp_out, sp_expression = signal_probability.sp_parse(sp_content)

	# work out sp for each node
	global node_order
	global sp_order
	node_order, sp_order = signal_probability.node_sp(primary_inputs, input_vec, cell_typ, wire_in_r, wire_out, cell_in_r, sp_typ, sp_out, sp_expression)

	# read stacking effect definition file
	stacking_content = general.read_txt(stacking_file)

	# parse stacking effect definition expression
	global stacking_typ_u
	global stacking_input_u
	global expression_u
	global stacking_typ_d
	global stacking_input_d
	global expression_d
	stacking_typ_u, stacking_input_u, expression_u, stacking_typ_d, stacking_input_d, expression_d = signal_probability.stacking_parse(stacking_content)

	# reoder cell types, input pins and instance names according to conenctions
	global cell_typ_order
	global cell_in_order
	global instance_order
	global node_order2
	cell_typ_order, cell_in_order, instance_order, node_order2 = signal_probability.others_reorder(node_order, cell_typ, cell_in_r, wire_in_r, instance_r)
	
	# sp mean and standard deviation
	print '\nInput vector done: calculating signal probabilities mean & deviation'
	global sp_mean
	sp_mean, sp_std = signal_probability.sp_mean_std(sp_order)

	# constraints to log file
	log = open(outdir+log_path, 'a')
	log.write('delay constraint = %f\narea constraint = %f\n' %(delay_constraint, area_constraint))

# *******************************************************************************


# *******************************************************************************
def evaluation(evaluation_cnt):
	
	print '--------------------------------------------------------------------------------------------------'	
	
	print 'Iteration %d: evaluate current design ...' %evaluation_cnt
	
	# run Design Compiler with tcl file in output directory
	print 'running Design Compiler in background ...'
	general.run_dc(outdir, tcl)

	print 'parsing area/timing/power/sdf report ...'

	# read DC report
	area_content = general.read_txt(outdir+area_rpt)	# area
	rise_timing_content = general.read_txt(outdir+rise_timing_rpt)	# timing
	fall_timing_content = general.read_txt(outdir+fall_timing_rpt)
	timing_content1 = rise_timing_content + fall_timing_content
	timing_content = [x.replace('<-', '') for x in timing_content1]
	sdf_content = general.read_txt(outdir+sdf_file)		# standard delay format

	# parse DC report
	area = area_parser.analyze(area_content)
	timing = timing_parser.analyze(timing_content)
	sdf = sdf_parser.analyze(sdf_content)

	# work out cell input to input delay
	new_delay, port_in = sdf_parser.in2in_delay(timing.instance, timing.delay, sdf.wire_in, sdf.wire_out, sdf.delay, sdf.wire_out_port)

	# stress duty cycles for all transistors of the critical paths
	nbti_path = signal_probability.find_alpha(port_in, instance_order, cell_in_order, cell_typ_order, sp_order, stacking_typ_u, stacking_input_u, expression_u, stacking_typ_d, stacking_input_d, expression_d, node_order, node_order2)

	# convert port name to wire name
	wire_in = ageing.port2wire(port_in, instance_r, cell_in_r, wire_in_r)

	# applying ageing model
	delta_mean, path_delay, _ = ageing.iteration(vec_no, nbti_path, wire_in, new_delay, timing.transition, timing.delay)
	
	print  '\nCurrent design status (Iteration %d): \narea: %f(um^2), max path delay: %f(ns), max degradation (normalized mean): %f' %(evaluation_cnt, area, max(path_delay), max(delta_mean))
	print '--------------------------------------------------------------------------------------------------'

	# current design status to log file
	log = open(outdir+log_path, 'a')
	log.write('--------------------------------------------------------------------------------\nIteration %d:\n' %evaluation_cnt)
	log.write('area = %f\n' %area)
	log.write('max path delay = %f\n' %max(path_delay))
	log.write('max degradation = %f\n' %max(delta_mean))
	if sensitivity_type == 'delay/area':
		x_decrease = 'delay_decrease'
	
	else:
		x_decrease = 'degradation_decrease'
	log.write('\t\tinstance\tcell\treplaced_cell\t%s\t\tarea_increase\t\tsensitivity\n' %x_decrease)

	return area, timing, delta_mean, path_delay

# *******************************************************************************


# *******************************************************************************
def trial(evaluation_cnt, area, timing, delta_mean, path_delay):

	path_i_delta = delta_mean.index(max(delta_mean))
	path_i_delay = path_delay.index(max(path_delay))

	# all instances for these two paths
	#instance_list = list(set(timing.instance[path_i_delta][1:]) | set(timing.instance[path_i_delay][1:]))
	all_instance_list = timing.instance[path_i_delay][1:]

	# select instances according to the maximum number of trial
	if len(all_instance_list) > max_trial:
		random.seed(trial_seed)
		instance_list = random.sample(all_instance_list, max_trial)
	else:
		instance_list = all_instance_list	

	instance_record = []
	old_cell_record = []
	new_cell_record = []
	area_record = []
	area_increase_record = []
	delay_decrease_record = []
	delta_decrease_record = []
	path_delay_record = []

	# sumarize trial
	instance_number = len(instance_list)
	print 'trial to upsize %d instances: %s' %(instance_number, ','.join(x for x in instance_list))

	# generate random
	random.seed(evaluation_cnt)
	random_value = [random.random() for i in range(instance_number)]

	# for all instance in current path
	for trial_i, instance in enumerate(instance_list):

		if random_value[trial_i] <= 1:

			# find path index in max delay/delta paths
			if instance in timing.instance[path_i_delay]:
				path_i = path_i_delay
			else:
				path_i = path_i_delta

			# find instance index in path
			instance_i = timing.instance[path_i].index(instance)
	
			# find cell with smallest size larger than old one
			opt_cell, _, max_size = lib_parser.smallest_bigger_cell(timing.cell_typ[path_i][instance_i], timing.cell_size[path_i][instance_i], lib.cell_typ, lib.cell_size)
			print '--------------------------------------------------------------------------------------------------'
			print 'Iteration %d, trial %d: upsizing instance %s, %s -> %s' %(evaluation_cnt, trial_i, timing.instance[path_i][instance_i], timing.cell_name[path_i][instance_i], opt_cell)

			# reset renamed netlist
			shutil.copy(outdir+netlist_backup, outdir+netlist_rename)
	
			# update renamed netlist
			optimization.update_netlist(outdir+netlist_rename, outdir+netlist_rename, timing.instance[path_i][instance_i], timing.cell_name[path_i][instance_i], opt_cell, max_size)

			# run Design Compiler with tcl file in output directory
			print 'running Design Compiler in background ...'
			general.run_dc(outdir, tcl)

			# read DC report
			area_content = general.read_txt(outdir+area_rpt)	# area
			rise_timing_content = general.read_txt(outdir+rise_timing_rpt)	# timing
			fall_timing_content = general.read_txt(outdir+fall_timing_rpt)
			timing_content1 = rise_timing_content + fall_timing_content
			timing_content = [x.replace('<-', '') for x in timing_content1]
			sdf_content = general.read_txt(outdir+sdf_file)		# standard delay format

			# parse DC report
			area_opt = area_parser.analyze(area_content)
			timing_opt = timing_parser.analyze(timing_content)
			sdf_opt = sdf_parser.analyze(sdf_content)

			# work out cell input to input delay
			new_delay_opt, port_in_opt = sdf_parser.in2in_delay(timing_opt.instance, timing_opt.delay, sdf_opt.wire_in, sdf_opt.wire_out, sdf_opt.delay, sdf_opt.wire_out_port)

			# stress duty cycles for all transistors of the critical paths
			nbti_path = signal_probability.find_alpha(port_in_opt, instance_order, cell_in_order, cell_typ_order, sp_order, stacking_typ_u, stacking_input_u, expression_u, stacking_typ_d, stacking_input_d, expression_d, node_order, node_order2)

			# convert port name to wire name
			wire_in_opt = ageing.port2wire(port_in_opt, instance_r, cell_in_r, wire_in_r)

			# applying ageing model
			delta_mean_opt, path_delay_opt, _ = ageing.iteration(vec_no, nbti_path, wire_in_opt, new_delay_opt, timing_opt.transition, timing_opt.delay)

			# parameter changes
			area_increase = float(area_opt) - float(area)
			delay_decrease = max(path_delay) - max(path_delay_opt)
			delta_decrease = max(delta_mean) - max(delta_mean_opt)

			# update record
			instance_record.append(timing.instance[path_i][instance_i])
			old_cell_record.append(timing.cell_name[path_i][instance_i])
			new_cell_record.append(opt_cell)
			area_record.append(area_opt)
			area_increase_record.append(area_increase)
			delay_decrease_record.append(delay_decrease)
			delta_decrease_record.append(delta_decrease)
			path_delay_record.append(max(path_delay_opt))

			print '\nIteration %d, trial %d/%d: area +(%f), max delay +(%f), max degredation +(%f)' %(evaluation_cnt, trial_i, instance_number-1, area_increase, -delay_decrease, -delta_decrease)
			print '--------------------------------------------------------------------------------------------------'

			# current trial status to log file
			log = open(outdir+log_path, 'a')
			log.write('trial %d\t\t' %trial_i)
			log.write('%s\t' %timing.instance[path_i][instance_i])
			log.write('%s\t' %timing.cell_name[path_i][instance_i])
			log.write('%s\t' %opt_cell)
			if sensitivity_type == 'delay/area':
				x_decrease = delay_decrease
				y_decrease = delta_decrease
			else:
				x_decrease = delta_decrease
				y_decrease = delay_decrease
			log.write('%s\t\t' %x_decrease)
			log.write('%s\t\t' %area_increase)
			x_sens = [area_increase and x_decrease/area_increase or x_decrease*float('inf')]
			#log.write('%s\t' %x_sens[0])
			log.write('%s\t' %x_sens[0])
		
			y_sens = [area_increase and y_decrease/area_increase or y_decrease*float('inf')]
			#log.write('%s\t\n' %y_sens[0])
			log.write('%s\t\n' %y_decrease)
		
			# record delay/degradation changes
			#delay_record = open(outdir+delay_log_path, 'a')
			#delta_record = open(outdir+delta_log_path, 'a')
		
			#if instance_i == 1:
			#	delay_record.write('[%s' %delay_decrease)
			#	delta_record.write('[%s' %delta_decrease)
			#else:
			#	delay_record.write(',%s' %delay_decrease)
			#	delta_record.write(',%s' %delta_decrease)
		
		#delay_record.write('];\n')
		#delta_record.write('];\n')

	return instance_record, old_cell_record, new_cell_record, area_record, area_increase_record, delay_decrease_record, delta_decrease_record, path_delay_record

# *******************************************************************************


# *******************************************************************************
def decision(evaluation_cnt, instance_record, old_cell_record, new_cell_record, area_record, area_increase_record, delay_decrease_record, delta_decrease_record, path_delay, path_delay_record):

	if sensitivity_type == 'delay/area':
		print 'sensitivity type: delay reduction/area increase'
		x_decrease_record = delay_decrease_record
		y_decrease_record = delta_decrease_record
	
	else:
		print 'sensitivity type: degradation redcution/area increase'
		x_decrease_record = delta_decrease_record
		y_decrease_record = delay_decrease_record
		
	best_index, optimization_end_area, sensitivity = optimization.sensitivity(x_decrease_record, y_decrease_record, area_increase_record, area_record, area_constraint, area0, delay_constraint, delay0, seed_2, path_delay_record)

	if optimization_end_area == True:
		print 'Opitmization complete: no transformation to satisfy area constraint' 
	else:
		best_instance = instance_record[best_index]
		best_old_cell = old_cell_record[best_index]
		best_new_cell = new_cell_record[best_index]

		print 'Iteration %d: decision made to upsize instance %s, %s -> %s' %(evaluation_cnt, best_instance, best_old_cell, best_new_cell)
		print '--------------------------------------------------------------------------------------------------'
		
		# update backup netlist
		optimization.update_netlist(outdir+netlist_backup, outdir+netlist_backup, best_instance, best_old_cell, best_new_cell, False)
			
		# update renamed netlist
		shutil.copy(outdir+netlist_backup, outdir+netlist_rename)

	# current decision to log file
	log = open(outdir+log_path, 'a')
	log.write('decision made\t')
	log.write('%s\t' %best_instance)
	log.write('%s\t' %best_old_cell)
	log.write('%s\t' %best_new_cell)
	log.write('%s\t\t' %x_decrease_record[best_index])
	log.write('%s\t\t' %area_increase_record[best_index])
	x_sens = [area_increase_record[best_index] and x_decrease_record[best_index]/area_increase_record[best_index] or x_decrease_record[best_index]*float('inf')]
	log.write('%s\t\n' %x_sens[0])
	log.write('area+ = %f\n' %area_increase_record[best_index])
	log.write('delay- = %f\n' %delay_decrease_record[best_index])
	log.write('degradation- = %f\n\n' %delta_decrease_record[best_index])

	# if max delay smaller than constraint
	if max(path_delay) - delay_decrease_record[best_index] <= delay_constraint:		
		optimization_end_delay = True
		print 'Opitmization complete: satisfy delay constraint'
	else:
		optimization_end_delay = False

	optimization_end = optimization_end_area or optimization_end_delay
	return optimization_end

# *******************************************************************************


# *******************************************************************************
def optimize(evaluation_cnt):

	area, timing, delta_mean, path_delay = evaluation(evaluation_cnt)
	
	if evaluation_cnt == 0:
		global area0
		area0 = area
		
		global delay0
		delay0 = max(path_delay)

	instance_record, old_cell_record, new_cell_record, area_record, area_increase_record, delay_decrease_record, delta_decrease_record, path_delay_record = trial(evaluation_cnt, area, timing, delta_mean, path_delay)
	optimization_end = decision(evaluation_cnt, instance_record, old_cell_record, new_cell_record, area_record, area_increase_record, delay_decrease_record, delta_decrease_record, path_delay, path_delay_record)

	if optimization_end == True:
		_, _, _, _ = evaluation(evaluation_cnt+1)
	
	return optimization_end

# *******************************************************************************


# *******************************************************************************
def main():	

	cnt = 0
	optimization_end = False

	prepare()

	print '**************************************************************************************************'
	start_time = time.time()
	while optimization_end == False:
		optimization_end = optimize(cnt)
		cnt = cnt + 1

	print '**************************************************************************************************'
	print 'Optimization takes %s seconds' %(time.time()-start_time)

# *******************************************************************************

main()
