import general, timing_parser

# *******************************************************************************
# analyze DC SDF file (content = content.readlines())
class analyze:
	
	def __init__(self, content):
		block = self.interconnect(content)
		self.wire_in, self.wire_out, self.delay, self.wire_out_port = self.wire_info(block)
	
	# find interconnect block
	def interconnect(self, content):
		block = []
		for line in content:
			if 'INTERCONNECT' in line:
				block.append(line)
		return block

	# interconnect information
	def wire_info(self, block):

		wire_in = []	# left node
		wire_out = []	# right node
		delay = []	# interconnect delay
		wire_out_port = []	# right port name	
		
		# for each interconnect
		for wire in block:
			wire_space = wire.split(None,4)
			wire_in.append(wire_space[1].split('/',1)[0].replace("\\", ""))
			wire_out.append(wire_space[2].split('/',1)[0].replace("\\", ""))
			delay.append(wire_space[3].split(':',2)[1])
			wire_out_port.append(wire_space[2])
		
		return wire_in, wire_out, delay, wire_out_port
	
# *******************************************************************************


# *******************************************************************************
# find out cell input to input delay
def in2in_delay(instance, cell_delay, wire_in, wire_out, wire_delay, wire_out_port):

	in2in_delay = []
	port_in = []

	# for each path 
	for path_index, path in enumerate(instance):
		cell_index = 1
		in2in_delay.append([])
		port_in.append([])

		# for each instance of the path
		for c1, c2 in zip(path[:-1],path[1:]):
			
			# wire delay between out_1 to in_2
			all_in_index = [i for i, ltr in enumerate(wire_in) if ltr == c1]
			all_out_index = [i for i, ltr in enumerate(wire_out) if ltr == c2]
			wire1_index = list(set(all_in_index) & set(all_out_index))[0]			
			D_wire1 = float(wire_delay[wire1_index])
			port_in[path_index].append(wire_out_port[wire1_index])
			
			# wire delay between out_2 to in_3
			if c2 in wire_in:
				wire2_index = wire_in.index(c2)
				D_wire2 = float(wire_delay[wire2_index])
			else:
				D_wire2 = 0			
			
			out2out_delay = float(cell_delay[path_index][cell_index])
			in2in_delay[path_index].append(out2out_delay-D_wire1+D_wire2)
			cell_index = cell_index + 1
		
	return in2in_delay, port_in

# *******************************************************************************


# *******************************************************************************
# test program
def test():
	sdf_path = '../dc_syn/test_opt.sdf'

	sdf = general.read_txt(sdf_path)
	sdf_info = analyze(sdf)
	print sdf_info.wire_in
	print sdf_info.wire_out
	print sdf_info.delay
	
	rpt_path = '../dc_syn/test_timing.rpt'

	rpt = general.read_txt(rpt_path)
	rpt_info = timing_parser.analyze(rpt)
	new_delay = in2in_delay(rpt_info.instance, rpt_info.delay, sdf_info.wire_in, sdf_info.wire_out, sdf_info.delay)
	print new_delay

#test()

# *******************************************************************************
