import random, math_parser, general, netlist_parser, lib_parser, math
from sys import stdout

# *******************************************************************************
# generate random input signal probabilities
def generate(number_of_input, iteration, seed):
	random.seed(seed)
	random_input = [[random.random() for i in range(number_of_input)] for j in range(iteration)]
	#random_input = [[0.5 for i in range(number_of_input)] for j in range(iteration)]
	return random_input

# *******************************************************************************


# *******************************************************************************
# parse sp file
def sp_parse(content):
	
	cell_typ = []
	pin_out = []
	expression = []

	for i, line in enumerate(content):

		line_split = line.split()

		# cell type
		cell_typ.append(line_split[0])

		pin_out.append([])
		expression.append([])

		for value in range(1,len(line_split)):
		
			out_split = line_split[value].split('=')			
			
			# output pin
			pin_out[i].append(out_split[0])

			# output sp expression
			expression[i].append(out_split[1])

	return cell_typ, pin_out, expression		
		
# *******************************************************************************


# *******************************************************************************
# parse stacking effect file
def stacking_parse(content):
	
	# pull-up network
	cell_typ_up = []
	stacking_input_up = []
	expression_up = []
	
	# pull-down network
	cell_typ_down = []
	stacking_input_down = []
	expression_down = []

	for index, line in enumerate(content):
		
		line_split = line.split()
		
		# pull-up network
		if line_split[0] == 'r':
			
			cell_typ_up.append(line_split[1]) # cell type
			stacking_input_up.append(line_split[2]) # stacking input
			expression_up.append(line_split[3]) # expression

		# pull-down network
		elif line_split[0] == 'f':
			
			cell_typ_down.append(line_split[1]) # cell type
			stacking_input_down.append(line_split[2]) # stacking input
			expression_down.append(line_split[3]) # expression
	
	return cell_typ_up, stacking_input_up, expression_up, cell_typ_down, stacking_input_down, expression_down		
		
# *******************************************************************************


# *******************************************************************************
# find selected cell information
def find_cell(cell_typ, all_typ, all_out, all_expression):
	if cell_typ not in all_typ:
		print '\ncell %s is not defined in sp file' %cell_typ
	index = all_typ.index(cell_typ)
	cell_out = all_out[index]
	cell_expression = all_expression[index]
	
	# cell_out: selected cell output
	# cell_expression: selected cell sp expression
	return cell_out, cell_expression

# *******************************************************************************


# *******************************************************************************
# work out output signal probabilities
def spout(expression, inputs, sp_in):

	input_def = '{'
	
	# expression input definition
	for i in range(0, len(inputs)):
		input_def = input_def + '\'' + inputs[i] + '\'' + ':' + str(sp_in[i]) + ','

	input_def = input_def[:-1] + '}'
	
	if len(inputs) == 1:
		sp_out = math_parser.evaluate(expression, {inputs[0]:sp_in[0]})
	elif len(inputs) == 2:
		sp_out = math_parser.evaluate(expression, {inputs[0]:sp_in[0], inputs[1]:sp_in[1]})
	elif len(inputs) == 3:
		sp_out = math_parser.evaluate(expression, {inputs[0]:sp_in[0], inputs[1]:sp_in[1], inputs[2]:sp_in[2]})
	elif len(inputs) == 4:
		sp_out = math_parser.evaluate(expression, {inputs[0]:sp_in[0], inputs[1]:sp_in[1], inputs[2]:sp_in[2], inputs[3]:sp_in[3]})
	elif len(inputs) == 5:
		sp_out = math_parser.evaluate(expression, {inputs[0]:sp_in[0], inputs[1]:sp_in[1], inputs[2]:sp_in[2], inputs[3]:sp_in[3], inputs[4]:sp_in[4]})
	elif len(inputs) == 6:
		sp_out = math_parser.evaluate(expression, {inputs[0]:sp_in[0], inputs[1]:sp_in[1], inputs[2]:sp_in[2], inputs[3]:sp_in[3], inputs[4]:sp_in[4], inputs[5]:sp_in[5]})
	else:
		print 'The number of inputs are not defined yet!'	

	return sp_out	

# *******************************************************************************


# *******************************************************************************
# produce node signal probilities
def node_sp(primary_in, sp_in, cell_typ_r, wire_in_r, wire_out_r, cell_in_r, all_typ, all_out, all_expression):

	# node order
	node_order = primary_in
	for i, cell in enumerate(cell_typ_r):
		if wire_out_r[i] not in node_order:
			node_order.extend(wire_out_r[i])

	# sp for each node
	sp_order = sp_in
	for iteration, primary_sp in enumerate(sp_in):

		stdout.write('\rInput vector %d/%d: calculating node signal probabilities' %((iteration+1), len(sp_in)))
		stdout.flush()
		
		# for each cell
		for i, cell in enumerate(cell_typ_r):
	
			# find cell sp expression
			_, expression = find_cell(cell, all_typ, all_out, all_expression)
		
			# cell inputs
			inputs = cell_in_r[i]

			# cell input sp
			cell_sp_in = []
			for i_port, in_port in enumerate(cell_in_r[i]):
				node_index = node_order.index(wire_in_r[i][i_port])
				cell_sp_in.append(sp_order[iteration][node_index])
			
			if len(expression) == len(wire_out_r[i]):

				# for each output
				for j, output in enumerate(wire_out_r[i]):
					sp_out = spout(expression[j], inputs, cell_sp_in)
					sp_order[iteration].append(sp_out)		

			else:
				print 'netlist doesn\'t not match sp file'
	
	return node_order, sp_order	
		
# *******************************************************************************


# *******************************************************************************
# reorder cell type, cell in and instance name
def others_reorder(node_order, cell_typ_r, cell_in_r, wire_in_r, instance_r):

	cell_typ_reshape = []
	cell_in_reshape = []
	wire_in_reshape = []
	instance_reshape = []

	for index_cell, each_cell in enumerate(wire_in_r):
		for index_wire, each_wire in enumerate(each_cell):

			# reshape cell_typ
			cell_typ_reshape.append(cell_typ_r[index_cell])

			# reshape cell_in
			cell_in_reshape.append(cell_in_r[index_cell][index_wire])

			# reshpae wire_in
			wire_in_reshape.append(each_wire)

			# reshape instance
			instance_reshape.append(instance_r[index_cell])

	
	cell_typ_order = []
	cell_in_order = []
	instance_order = []
	node_order2 = []

	# for each node
	for node in node_order:
		cell_typ = [x for i, x in enumerate(cell_typ_reshape) if wire_in_reshape[i] == node]
		cell_in = [x for i, x in enumerate(cell_in_reshape) if wire_in_reshape[i] == node]
		instance = [x.replace('\\','') for i, x in enumerate(instance_reshape) if wire_in_reshape[i] == node]
		node = [x for i, x in enumerate(wire_in_reshape) if wire_in_reshape[i] == node]

		cell_typ_order.extend(cell_typ)
		cell_in_order.extend(cell_in)
		instance_order.extend(instance)
		node_order2.extend(node)

	return cell_typ_order, cell_in_order, instance_order, node_order2			

# *******************************************************************************


# *******************************************************************************
# work out stress duty cycles for all cells of the path
def find_alpha(port_in, instance_order, cell_in_order, cell_typ_order, sp_order, cell_typ_up, stacking_input_up, expression_up, cell_typ_down, stacking_input_down, expression_down, node_order, node_order2):

	nbti_path = []
	pbti_path = []

	# for each path
	for path_i, path in enumerate(port_in):
	
		nbti_path.append([])
		pbti_path.append([])

		# for each port
		for port in path:
			instance = port.split('/')[0]
			inport = port.split('/')[1]

			# instance plus port index in reordered netlist
			port_index = zip(instance_order, cell_in_order).index((instance, inport))
			cell_typ = cell_typ_order[port_index]

			# check if there is stacking effect
			# pull-up network
			if (cell_typ, inport) in zip(cell_typ_up, stacking_input_up):
				
				# find stacking expression
				expression_index = zip(cell_typ_up, stacking_input_up).index((cell_typ, inport))
				expression = expression_up[expression_index]		

				# find all input pins related to stacking expression
				all_input = [letter for letter in expression if str.isalpha(letter) == True]
				
				# find sp for each input pin
				all_sp_index = []
				for pin in all_input:
					sp_index2 = zip(instance_order, cell_in_order).index((instance, pin))
					sp_index = node_order.index(node_order2[sp_index2])
					all_sp_index.append(sp_index)
			
				# work out stress duty cycle
				nbti_alpha = []
				for sp_group in sp_order:
					stacking_up_sp = [sp_group[i] for i in all_sp_index]
					nbti_row = spout(expression, all_input, stacking_up_sp)
					nbti_alpha.append(nbti_row)
				
				nbti_path[path_i].append(nbti_alpha)
					
			# pull-down network
			#elif (cell_typ, inport) in zip(cell_typ_down, stacking_input_down):
			#	print 'stacking pull-down'
			
			# no stacking effect
			else:
				sp_index2 = zip(instance_order, cell_in_order).index((instance, inport))
				sp_index = node_order.index(node_order2[sp_index2])
				nbti_alpha = [1-sp_group[sp_index] for sp_group in sp_order]
				pbti_alpha = [sp_group[sp_index] for sp_group in sp_order]
				nbti_path[path_i].append(nbti_alpha)
				pbti_path[path_i].append(pbti_alpha)

	return nbti_path
	

# *******************************************************************************
	

# *******************************************************************************
# sp mean and std
def sp_mean_std(sp_order):

	sp_rotate = zip(*sp_order)
	sp_std = []
	
	# sp mean of each node
	sp_mean = [sum(sp)/len(sp) for sp in sp_rotate]
	
	# sp variance
	for i, mean in enumerate(sp_mean):
		variance = sum((x - mean)**2 for x in sp_rotate[i])/len(sp_rotate[i])
		
	# sp standard deviation
		sp_std.append(math.sqrt(variance))

	return sp_mean, sp_std

# *******************************************************************************


# *******************************************************************************
# test program
def test():
	random_input = generate(16, 1, 123)
	print random_input
	
	sp_file = '../verilog/st65.sp'
	sp = general.read_txt(sp_file)
	all_typ, all_out, all_expression = sp_parse(sp)
	cell_out, cell_expression = find_cell('HS65_LS_IVX', all_typ, all_out, all_expression)
	print cell_out
	print cell_expression
	
	sp_out = spout(cell_expression[0], ['A'], [0.3])

	netlist_path = '../dc_syn/test.v'
	netlist = general.read_txt(netlist_path)

	library = '../verilog/CORE65LPSVT_bc_1.05V_105C.lib'
	lib_content = general.read_txt(library)
	lib = lib_parser.analyze(lib_content)
	
	netlist_info = netlist_parser.analyze(netlist, lib)
	instance_r, cell_typ_r, wire_in_r, wire_out_r, cell_in_r, cell_out_r = netlist_parser.reorder(netlist_info.instance, netlist_info.cell_typ, netlist_info.wire_in, netlist_info.wire_out, netlist_info.cell_in, netlist_info.cell_out, netlist_info.inputs)
	
	node_order, sp_order = node_sp(netlist_info.inputs, random_input, cell_typ_r, wire_in_r, wire_out_r, cell_in_r, all_typ, all_out, all_expression)
	
	node_list = ['B[0]','n40', 'n41', 'n42', 'n49', 'n21', 'n50', 'n51', 'n58', 'n18', 'n59', 'n60', 'n67', 'n15', 'n68', 'n69', 'n76', 'n12', 'n77', 'n78', 'n88', 'n9', 'n89', 'n90', 'n92', 'n93', 'n3']
	
	for node in node_list:
		index_node = node_order.index(node)
	
		sp_all = []
		for i in range(0,1):
			sp_all.append(sp_order[i][index_node])
	
		sp_mean = sum(sp_all)/1
		print sp_mean

	print node_order
	print sp_order
	
	
#test()

# *******************************************************************************	
