import general, re

# ******************************************************************************
# analyze DC timing report (content = content.readlines())
class analyze:

	def __init__(self, content):
		block, self.start, self.end = self.path_block(content)
		self.instance, self.cell_name, self.delay, self.transition, self.port_name = self.cell_info(block)
		self.cell_typ, self.cell_size = self.split_typ_size(self.cell_name)
	

	# find path blocks
	def path_block(self, content):
		# find start/end lines of path block
		start = []
		end = []
		for number, line in enumerate(content):
			if 'input external delay' in line:
				start.append(number)
			if 'data arrival time' in line and '-' not in line:
				end.append(number)
		# find path blocks
		block = []
		if len(start) != len(end):
			print 'Error when extract blocks'
		else:
			for index, number in enumerate(start):
				block.append([])
				block[index].append(content[number+1:end[index]-1])
		return block, start, end

	
	# cell information
	def cell_info(self, block):

		instance = []	# instance name
		cell_name = []	# cell name
		delay = []	# cell delay
		transition = []	# type of transition
		port_name = []	# output name

		# for each path of the blocks
		for index, path in enumerate(block):
			instance.append([])
			cell_name.append([])
			delay.append([])
			transition.append([])
			port_name.append([])

			# for each cell of the path
			for cell in path[0]:
				instance[index].append(cell.split(None,2)[0].split('/',1)[0].replace(' ',''))
				cell_name[index].append(cell[cell.find('(')+1:cell.find(')')])
				delay[index].append(cell.split(None,5)[2])
				transition[index].append(cell.split(None,5)[4])
				port_name[index].append(cell.split(None,2)[0])

		return instance, cell_name, delay, transition, port_name
	
	# split cell type and size
	def split_typ_size(self, cell_name):
		gate_size = []
		gate_typ = []

		# for each path	
		for i, path in enumerate(cell_name):
		
			gate_size.append([])
			gate_size[i].append([])

			gate_typ.append([])
			gate_typ[i].append([])
		
			# for each cell except the input 
			for cell in path[1:]:
			
				size = re.findall('\d+', cell)
			
				cnt = len(size[-1])
				gate_size[i].append(int(size[-1]))
				gate_typ[i].append(cell[:-cnt])
	
		return gate_typ, gate_size
	
# ******************************************************************************


# *******************************************************************************
# test program
def test():
	rpt_path = '../dc_syn/test_timing.rpt'

	rpt = general.read_txt(rpt_path)
	rpt_info = analyze(rpt)
	print rpt_info.cell_name
	print rpt_info.cell_typ
	print rpt_info.cell_size
	print rpt_info.delay

#test()

# *******************************************************************************
