import general, glob, lib_parser, ageing
from sys import stdout

# *******************************************************************************
# generate DC command for delay calculation
def dc_command(port_in, port_out, delay_path, min_lifetime_index):

	command = ''
	record = []
	
	# find path with minimum lifetime
	path = port_in[min_lifetime_index]

		
	# for each port in this path
	for cell_i, cell_in in enumerate(path):

		cell_out = port_out[min_lifetime_index][cell_i+1]
		combination = '%s-%s' %(cell_in, cell_out)

		instance = cell_in.split('/')[0]
		inport = cell_in.split('/')[1]
		outport = cell_out.split('/')[1]

		path = delay_path.replace('/', '')		
			
		# if cell delay has not been calculated
		if combination not in record:

			#combination.replace('/', '-')
			string = 'report_delay_calculation -from %s -to %s > %s/%s-%s-%s\n' %(cell_in, cell_out, path, instance, inport, outport)	
			record.append(combination)
			command = command + string
	return command

# *******************************************************************************


# *******************************************************************************
# replace commands
def replace(tcl, command):
	content = general.read_txt(tcl)
	content_new = []
	
	# remove all commands for delay calculation
	for line in content:
		if 'report_delay_calculation' not in line:
			content_new.append(line)

	# add new commands into tcl
	content_new.insert(len(content_new)-1, command)

	# update new tcl
	f = open(tcl, 'w')
	for line in content_new:
		f.write(line)
	f.close()			

# *******************************************************************************


# *******************************************************************************
# parse delay calculation report
class analyze:

	def __init__(self, result_path, lib, timing, min_lifetime_index, port_in, node_order, sp_mean, wire_name, delta_mean, path_delay, lifetime_mean, max_delay, max_area, current_area):
		instance, inport, outport, rise_cap, fall_cap, rise_effort, fall_effort = self.read(result_path)
		delta_stage_effort, delta_cap, out_cap, stage_effort, path_sp, delta_area = self.info_update(lib, timing, min_lifetime_index, port_in, instance, rise_cap, fall_cap, rise_effort, fall_effort, node_order, sp_mean, wire_name)
		self.sens_index = self.evaluation(delta_stage_effort, delta_cap, out_cap, stage_effort, path_sp, delta_mean, path_delay, lifetime_mean, max_delay, min_lifetime_index, delta_area, max_area, current_area)
	

	# read DC delay calculation reports
	def read(self, result_path):

		files = glob.glob(result_path + '/*')
		
		instance_list = []
		inport_list = []
		outport_list = []
		rise_cap_list = []
		fall_cap_list = []
		rise_effort_list = []
		fall_effort_list = []		

		# read all files in result path
		for i, result in enumerate(files):
			
			stdout.write('\rparsing delay calculation report %d/%d' %(i+1, len(files)))
			stdout.flush()

			content = general.read_txt(result)
			instance, inport, outport, rise_cap, fall_cap, rise_effort, fall_effort = self.parse(content)
			
			instance_list.append(instance)
			inport_list.append(inport)
			outport_list.append(outport)
			rise_cap_list.append(rise_cap)
			fall_cap_list.append(fall_cap)
			rise_effort_list.append(rise_effort)
			fall_effort_list.append(fall_effort)

		return 	instance_list, inport_list, outport_list, rise_cap_list, fall_cap_list, rise_effort_list, fall_effort_list 	

	
	# parse DC delay calculation reports
	def parse(self, content):

		# parse each line
		for index, line in enumerate(content):	
	
			# find instance and input definition
			if 'From pin:' in line:
				inport_index = index

			# find output definition
			if 'To pin:' in line:
				outport_index = index
		
			# find Rise Delay definition
			if 'Rise Delay' in line:
				rise_index = index

			# find Fall Delay definition
			if 'Fall Delay' in line:
				fall_index = index

		# cell instance
		no_space_line = ''.join(content[inport_index].split())
		instance_and_inport = no_space_line.split(':')[1]
		instance = instance_and_inport.split('/')[0]
	
		# inport name
		inport = instance_and_inport.split('/')[1]
	
		# outport name
		no_space_line = ''.join(content[outport_index].split())
		outport = no_space_line.split('/')[1]

		# rise capacitance
		no_space_line = ''.join(content[rise_index+5].split())
		rise_cap = no_space_line.split('=')[1]

		# fall capacitance
		no_space_line = ''.join(content[fall_index+5].split())
		fall_cap = no_space_line.split('=')[1]

		# rise delay stage effort
		no_space_line = ''.join(content[rise_index+13].split())
		rise_effort = no_space_line.split('=')[1].split('D')[0]

		# fall delay stage effort
		no_space_line = ''.join(content[fall_index+13].split())
		fall_effort = no_space_line.split('=')[1].split('D')[0]

		return instance, inport, outport, rise_cap, fall_cap, rise_effort, fall_effort


	# capacitance, stage effort change for upsizing each cell in the path
	def info_update(self, lib, timing, min_lifetime_index, port_in, instance, rise_cap, fall_cap, rise_effort, fall_effort, node_order, sp_mean, wire_name):
		
		print '\nupsizing trail(s) for instance(s): %s' %(','.join(x for x in timing.instance[min_lifetime_index][1:]))
		
		cell_instance = timing.instance[min_lifetime_index][1:]
		cell_name = timing.cell_name[min_lifetime_index][1:]
		cell_typ = timing.cell_typ[min_lifetime_index][1:]
		cell_size = timing.cell_size[min_lifetime_index][1:]
		cell_transition = timing.transition[min_lifetime_index][1:]
		cell_in = [string.split('/')[1] for string in port_in[min_lifetime_index]]

		wire_in = []
		wire_in.append([])
		for wire in wire_name[min_lifetime_index]:
			wire_in[0].append(wire)
		
		# find smallest cell than bigger than each cell in the path
		cell_opt_size = []
		delta_stage_effort = []
		delta_cap = []
		out_cap = []
		stage_effort = []
		path_sp = []
		delta_area = []

		index = 0

		for typ, size in zip(cell_typ, cell_size):
			
			cell_opt, opt_size, _ = lib_parser.smallest_bigger_cell(typ, size, lib.cell_typ, lib.cell_size)
			cell_opt_size.append(opt_size)

			# stage effort decrease
			delta_se = 1 - float(size)/float(opt_size)
			delta_stage_effort.append(delta_se)
			
			# output capacitances and changes and stage efforts
			# changes:
			cell_in_0, _, area_0, rise_cap_0, fall_cap_0 = lib_parser.find_cell_name(cell_name[index], lib.cell_name, lib.pin_in, lib.pin_out, lib.area, lib.rise_max_cap, lib.fall_max_cap)
			_, _, area_1, rise_cap_1, fall_cap_1 = lib_parser.find_cell_name(cell_opt, lib.cell_name, lib.pin_in, lib.pin_out, lib.area, lib.rise_max_cap, lib.fall_max_cap)
			
			port_index = cell_in_0.index(cell_in[index])

			# output capacitances and stage efforts:
			instance_index = instance.index(cell_instance[index])				

			if cell_transition[index] is 'r':
				delta_c = rise_cap_1[port_index] - rise_cap_0[port_index]
				out_c = rise_cap[instance_index]
				stage_e = rise_effort[instance_index]
			else:
				delta_c = fall_cap_1[port_index] - fall_cap_0[port_index]
				out_c = fall_cap[instance_index]
				stage_e = fall_effort[instance_index]

			delta_cap.append(delta_c)
			out_cap.append(out_c)
			stage_effort.append(stage_e)

			# area changes:
			delta_area.append(float(area_1) - float(area_0))	

			index = index + 1

		# signal probabilities
		path_sp = ageing.find_sp(node_order, sp_mean, wire_in, 0)

		return delta_stage_effort, delta_cap, out_cap, stage_effort, path_sp, delta_area


	# evaluate lifetime changes
	def evaluation(self, delta_stage_effort, delta_cap, out_cap, stage_effort, path_sp, delta_mean, path_delay, lifetime_mean, max_delay, min_lifetime_index, delta_area, max_area, current_area):
		
		delta_path_delay = []
		delta_nbti = []
		
		# delay change by upsizing 1st instance
		delta_path_d = -delta_stage_effort[0] * float(stage_effort[0]) * float(out_cap[0])
		delta_path_delay.append(delta_path_d)
		
		# nbti degradation change by upsizing 1st instance
		delta_n = ageing.ageing(path_sp[0][0], delta_path_d)
		delta_nbti.append(delta_n)

		# delay/nbti changes by upsizing the following instances
		for i in range(1, len(out_cap)):
			delta_path_d_0 = float(stage_effort[i-1]) * delta_cap[i-1]
			delta_path_d_1 = -delta_stage_effort[i] * float(stage_effort[i]) * float(out_cap[i]) 
			delta_path_delay.append(delta_path_d_0+delta_path_d_1)

			delta_n_0 = ageing.ageing(path_sp[0][i-1], delta_path_d_0) 
			delta_n_1 = ageing.ageing(path_sp[0][i], delta_path_d_1)
			delta_nbti.append(delta_n_0 + delta_n_1)
		
		# delay margin
		margin_0 = max_delay - path_delay[min_lifetime_index]
		margin = [margin_0 - x for x in delta_path_delay]

		# degradation
		nbti_0 = delta_mean[min_lifetime_index]
		nbti = [nbti_0 + x for x in delta_nbti]

		# lifetime
		lifetime = [x/y for x,y in zip(margin, nbti)]
		
		# lifetime changes
		lifetime_0 = lifetime_mean[min_lifetime_index]
		delta_lifetime = [x - lifetime_0 for x in lifetime]

		# sensitivity
		sensitivity = [y and x/y or x*float('inf') for x,y in zip(delta_lifetime, delta_area)]

		# sort sensitivity and return instance index
		sens_i = list(reversed(sorted(range(len(sensitivity)), key=lambda k:sensitivity[k])))
		
		# eliminate un-optimized decisions
		sens_index = []
		for i in sens_i:
			
			# remove negative sensitivities
			if sensitivity[i] > 0:
				
				# remove decision violating area constraint
				if float(current_area) + delta_area[i] <= float(max_area):
					sens_index.append(i)

		return sens_index

# *******************************************************************************
