import ageing, general, lib_parser, math_parser, netlist_parser, sdf_parser, signal_probability, timing_parser, re, os, shutil, optimization, area_parser, delay_calculation, time, sys

# *******************************************************************************
# ageing-aware synthesis main program
# definition:
# ----------------------------------------------------------------------------
# outdir - output directory name
# netlist_file - test netlist
# netlist_rename - renamed netlist
# netlist_backup - backup netlist
# tcl - Design Compiler tcl
# timing_rpt - DC timing report
# area_rpt - DC area report
# sdf_file - DC sdf
# lib_file - technology library file
# vec_no - number of input vectors
# seed - seed for random input vectors generation
# sp_file - signal probability expression file
# log_path - optimization log
# max_area - area constraint

# preparing
outdir = '../dc_syn'
netlist_file = '../lib/st65/test1.v'
netlist_rename = '/test.v'
netlist_backup = '/design_opt.v'
tcl = '../scripts/upsizing65.tcl'
vec_no = 2500
seed = 123
sp_file = '../lib/st65/st65.sp'
lib_file = '../lib/st65/CORE65LPSVT_bc_1.05V_105C.lib'
log_path = '/optimization.log'

# evaluation
timing_rpt = '/test_timing.rpt'
area_rpt = '/test_area.rpt'
sdf_file = '/test.sdf'
sample_path = '/sample.txt'

# *******************************************************************************
def prepare():		

	print 'ageing distribution analysis'

	# create output directory
	general.create_outdir(outdir)

	# move test netlist into output directory and rename
	shutil.copy(netlist_file, outdir+netlist_rename)

	# make a netlist backup
	shutil.copy(netlist_file, outdir+netlist_backup)
	
	print 'parsing library file ...'
	
	# read library file
	lib_content = general.read_txt(lib_file)

	# parse library file
	global lib
	lib = lib_parser.analyze(lib_content)
	
	print 'parsing netlist ...'

	# read netlist
	netlist_content = general.read_txt(outdir+netlist_rename)

	# parse netlist
	netlist = netlist_parser.analyze(netlist_content, lib)

	print 'parse netlist complete'

	# design primary inputs
	primary_inputs = netlist.inputs

	# reorder netlist according connections
	global instance_r
	global wire_in_r
	global cell_in_r
	instance_r, cell_typ, wire_in_r, wire_out, cell_in_r, cell_out = netlist_parser.reorder(netlist.instance, netlist.cell_typ, netlist.wire_in, netlist.wire_out, netlist.cell_in, netlist.cell_out, netlist.inputs)

	# generate input vectors
	input_vec = signal_probability.generate(len(primary_inputs), vec_no, seed)
	
	# read sp expression file
	sp_content = general.read_txt(sp_file)

	# parse sp expression
	sp_typ, sp_out, sp_expression = signal_probability.sp_parse(sp_content)

	# work out sp for each node
	global node_order
	global sp_order
	node_order, sp_order = signal_probability.node_sp(primary_inputs, input_vec, cell_typ, wire_in_r, wire_out, cell_in_r, sp_typ, sp_out, sp_expression)
	
	# sp mean and standard deviation
	print '\nInput vector done: calculating signal probabilities mean & deviation'
	global sp_mean
	sp_mean, sp_std = signal_probability.sp_mean_std(sp_order)
	
	print node_order
	print [100*(1-x) for x in sp_mean]

# *******************************************************************************


# *******************************************************************************
def evaluation():
	
	# run Design Compiler with tcl file in output directory
	print 'running Design Compiler in background ...'
	general.run_dc(outdir, tcl)

	print 'parsing area/timing/power/sdf report ...'

	# read DC report
	area_content = general.read_txt(outdir+area_rpt)	# area
	timing_content = general.read_txt(outdir+timing_rpt)	# timing
	sdf_content = general.read_txt(outdir+sdf_file)		# standard delay format

	# parse DC report
	area = area_parser.analyze(area_content)
	timing = timing_parser.analyze(timing_content)
	sdf = sdf_parser.analyze(sdf_content)

	# work out cell input to input delay
	new_delay, port_in = sdf_parser.in2in_delay(timing.instance, timing.delay, sdf.wire_in, sdf.wire_out, sdf.delay, sdf.wire_out_port)

	# convert port name to wire name
	wire_in = ageing.port2wire(port_in, instance_r, cell_in_r, wire_in_r)

	# applying ageing model
	delta_mean, _, delta = ageing.iteration(vec_no, node_order, sp_order, wire_in, new_delay, timing.transition, timing.delay)

	# record delay/degradation changes
	samples = open(outdir+sample_path, 'a')
	samples.write('[')
	
	for index, sample in enumerate(delta):
		if index != 0:
			samples.write(',')
		samples.write('%s' %sample[0])
		
	samples.write('];')

	print '\nanalysis complete, check sample file'

# *******************************************************************************


# *******************************************************************************
def main():
	
	prepare()
	evaluation()

# *******************************************************************************

main()
