import general, re

# *******************************************************************************
# find gate options in terms of size and port information
class analyze:
	
	def __init__(self, content):
		lines, number = self.find_cell(content)
		self.cell_typ, self.cell_size, self.cell_name = self.size(lines)
		#self.pin_in, self.pin_out, self.area, self.rise_max_cap, self.fall_max_cap = self.pin_area(content, number)
		self.pin_in, self.pin_out, self.area = self.pin_area(content, number)
	

	# find cell definition
	def find_cell(self, content):

		lines = []
		number = []

		for i, line in enumerate(content):
			
			no_space_line = ''.join(line.split())		
	
			if no_space_line[0:5] == 'cell(':
				lines.append(line)

				# beginning line for each cell
				number.append(i)
		number.append(len(content)-1)
		
		# lines: lines containing all cell definition
		return lines, number
	

	# size information
	def size(self, lines):
		cell_typ = []
		cell_size = []
		cell_name = []
		index = -1
		
		# for each line
		for line in lines:
			full_name = line[line.find('(')+1:line.find(')')]
		
			cell_name.append(full_name)		
					
			size = re.findall('\d+', full_name)
	
			# cell without size options
			if not size:
				if full_name not in cell_typ:
					cell_typ.append(full_name)
					index = index + 1
					cell_size.append([])
				cell_size[index].append(size)
			
			# cell with size options
			else:
				cnt = len(size[-1])
				if full_name[:-cnt] not in cell_typ:
					cell_typ.append(full_name[:-cnt])
					index = index + 1
					cell_size.append([])
					cell_size[index].append(size[-1])
				else:
					exist_index = cell_typ.index(full_name[:-cnt])
					cell_size[exist_index].append(size[-1])
		
		# cell_typ: all cell types
		# cell_size: all size options for each type
		return cell_typ, cell_size, cell_name

	
	# pin/area/capacitance information
	def pin_area(self, content, number):
	
		pin_in = []
		pin_out = []
		area = []
		rise_max_cap = []
		fall_max_cap = []
		
		for i in range(0, len(number)-1):
			pin_in.append([])
			pin_out.append([])
			rise_max_cap.append([])
			fall_max_cap.append([])
	
			rise_cap_all = []
			fall_cap_all = []
			pin = []
			pin_type = []
	
			for index, line in enumerate(content[number[i]:number[i+1]]):
				no_space_line = ''.join(line.split())

				if 'test_cell()' in no_space_line:
					break
					
				words = no_space_line.split(':',1)
		
				# for area definition
				if words[0] == 'area':
					area.append(float(words[1].replace(';', '')))

				# for pin definition
				if no_space_line[0:4] == 'pin(':
					pin.append(no_space_line[no_space_line.find('(')+1:no_space_line.find(')')])
		
				# for direction definition
				if words[0] == 'direction' and 'internal' not in no_space_line:
					pin_type.append(words[1].replace(';',''))

				# for capacitance definition
				#if 'rise_capacitance_range' in no_space_line or 'fall_capacitance_range' in no_space_line:
					
					# if max cap not in this line					
				#	if '\\' in no_space_line:
				#		cap_line = ''.join(content[number[i]:number[i+1]][index+1].split())
					# if max cap in this line
				#	else:
				#		cap_line = no_space_line					
				#	cap = re.findall("(?:\d+\.\d+|\d+)", cap_line)
		
					# rise capacitance
				#	if 'rise_capacitance_range' in no_space_line:
				#		rise_cap_all.append(float(cap[-1]))

					# fall capacitance
				#	else:
				#		fall_cap_all.append(float(cap[-1]))

			# input pin
			for index, each in enumerate(pin_type):
				if each == 'input':
					pin_in[i].append(pin[index])
				#	rise_max_cap[i].append(rise_cap_all[index])
				#	fall_max_cap[i].append(fall_cap_all[index])

			# output pin
			for index, each in enumerate(pin_type):
				if each == 'output':
					pin_out[i].append(pin[index])

		#return pin_in, pin_out, area, rise_max_cap, fall_max_cap
		return pin_in, pin_out, area


# *******************************************************************************


# *******************************************************************************
# find selected cell information based on cell name
def find_cell_name(cell_name, all_name, pin_in, pin_out, all_area, rise_cap, fall_cap):
	index = all_name.index(cell_name)	
	
	cell_in = pin_in[index]
	cell_out = pin_out[index]
	
	# selected cell area
	area = all_area[index]

	# selected cell capacitances
	cell_rise_cap = rise_cap[index]
	cell_fall_cap = fall_cap[index]
	
	return cell_in, cell_out, area, cell_rise_cap, cell_fall_cap

# *******************************************************************************


# *******************************************************************************
# find selected cell information based on cell type
def find_cell(cell_typ, all_typ, all_size, all_name, pin_in, pin_out):
	index = all_typ.index(cell_typ)
	cell_size = all_size[index]
	cell_example = cell_typ+cell_size[0]	
	
	port_index = all_name.index(cell_example)	
	
	cell_in = pin_in[port_index]
	cell_out = pin_out[port_index]
	
	# cell_size: all size options for selected cell type
	# cell_in: all inputs for selected cell type
	# cell_out: all outputs for selected cell type
	return cell_size, cell_in, cell_out

# *******************************************************************************


# *******************************************************************************
# find smallest size bigger than selected cell
def smallest_bigger_cell(cell_typ, cell_size, all_typ, all_size):
	index = all_typ.index(cell_typ)
	
	# convert size info of selected cell type from string into int 
	int_size = [int(value) for value in all_size[index]]
	
	# check if bigger size cell exists
	max_size = max(int_size) <= int(cell_size)
	
	if not max_size:
		# find the smallest cell bigger than given one
		opt_size = min(value for value in int_size if value > int(cell_size))
	else:
		opt_size = int(cell_size)
	
	# optimized cell type & size
	opt_cell = cell_typ + str(opt_size)
	
	return opt_cell, opt_size, max_size

# *******************************************************************************


# *******************************************************************************
# compare cell area
def area_compare(cell_name_1, cell_name_2, all_name, all_area):
	index_1 = all_name.index(cell_name_1)
	index_2 = all_name.index(cell_name_2)

	area_1 = all_area[index_1]
	area_2 = all_area[index_2]

	area_increase = float(area_2) - float(area_1)
	
	return area_increase

# *******************************************************************************


# *******************************************************************************
# modify lib to input transition independent
def lib2tin_independent(content, new_lib_path):

	fall_index = []
	rise_index = []

	modify_content = content

	for index, line in enumerate(content):
		
		no_space_line = ''.join(line.split())		
	
		if no_space_line[0:10] == 'cell_fall(':
			first_trans_line = content[index+1].replace('values(', '')

		elif no_space_line[0:10] == 'cell_rise(':
			first_trans_line = content[index+1].replace('values(', '')

		if (no_space_line[0:10] == 'cell_rise(' or no_space_line[0:10] == 'cell_fall(') and ';' not in first_trans_line:
			cnt = 2
			while ';' not in modify_content[index+cnt]:
				modify_content[index+cnt] = first_trans_line
				cnt = cnt + 1
			modify_content[index+cnt] = first_trans_line.replace(',\\', ');')

	modify_lib = open(new_lib_path, 'a')
	for line in modify_content:
		modify_lib.write(line)

# *******************************************************************************


# *******************************************************************************
# test program
def test():
	library = '../lib/st65/CORE65LPSVT_bc_1.05V_105C.lib'
	lib = general.read_txt(library)

	cell_info = analyze(lib)
	print cell_info.cell_typ[0:2]
	print cell_info.cell_size[0:2]
	print cell_info.cell_name[0:2]
	print cell_info.pin_in[0:2]
	print cell_info.pin_out[0:2]
	print cell_info.area[0:2]

	size_option, cell_in, cell_out = find_cell('HS65_LSS_XNOR2X', cell_info.cell_typ, cell_info.cell_size, cell_info.cell_name, cell_info.pin_in, cell_info.pin_out)
	print size_option
	print cell_in
	print cell_out

	new_lib_path = '../lib/st65/CORE65LPSVT_bc_1.05V_105C_modify.lib'
	lib2tin_independent(lib, new_lib_path)

	#opt_cell, max_size = smallest_bigger_cell('HS65_LSS_XNOR2X', 12, cell_info.cell_typ, cell_info.cell_size)
	#print opt_cell, max_size

#test()

# *******************************************************************************

