import general, shutil, lib_parser, netlist_parser, signal_probability, timing_parser, sdf_parser, ageing, area_parser, tree_matching, time

outdir = '../dc_syn_b'
netlist_file = '../lib/st65/iscas85/c499_syn.v'
netlist_rename = '/test.v'
tcl = '../scripts/mapping_test.tcl'
vec_no = 1
seed = 123
timing_cons = 2.375
looser_timing = 0.5
sp_file = '../lib/st65/st65.sp'
lib_file = '../lib/st65/CORE65LPSVT_nom_1.10V_25C.lib'
rise_timing_rpt = '/test_timing_rise.rpt'
fall_timing_rpt = '/test_timing_fall.rpt'
area_rpt = '/test_area.rpt'
sdf_file = '/test.sdf'
new_timing_rpt = '/post_ageing.rpt'
stacking_file = '../lib/st65/st65_stacking.txt'
logic_file = '../lib/st65/st65_logic.txt'
min_delay_tree_netlist = 'min_delay_tree.v'
min_deg_tree_netlist = 'min_deg_tree.v'
tree_tcl = '../scripts/tree_syn.tcl'
backward_tcl = '../scripts/backward.tcl'

# In order the estimate the degradation, disable remap_enable (remap_enable = False)
# To perform logic restructuring, enable remap_enable (remap_enable = True) 
remap_enable = False

min_area_enable = True
max_trial = 5
delta_log_path = '/c7552_opt_delta.dat'
delta_log_mode = 'latex'
netlist_r_text = '../lib/st65/EPFL/log2_reorder.txt'

# *******************************************************************************
def prepare():		

	print 'ageing-aware technology mapping...'

	# create output directory
	general.create_outdir(outdir)

	# move test netlist into output directory and rename
	shutil.copy(netlist_file, outdir+netlist_rename)
	
	print 'parsing library file ...'
	
	# read library file
	lib_content = general.read_txt(lib_file)

	# parse library file
	global lib
	lib = lib_parser.analyze(lib_content)
	
	print 'parsing netlist ...'

	# read netlist
	netlist_content = general.read_txt(outdir+netlist_rename)

	# remove '\.../' from netlist
	netlist_parser.virgule_removal(netlist_content, outdir+netlist_rename)

	# read netlist
	netlist_content = general.read_txt(outdir+netlist_rename)

	# re-read netlist
	netlist_content = general.read_txt(outdir+netlist_rename)

	# parse netlist
	global netlist
	netlist = netlist_parser.analyze(netlist_content, lib)

	print 'parse netlist complete'

	# design primary inputs
	primary_inputs = netlist.inputs

	# reorder netlist according connections
	global instance_r
	global wire_in_r
	global cell_in_r
	instance_r, cell_typ, wire_in_r, wire_out, cell_in_r, cell_out = netlist_parser.reorder(netlist.instance, netlist.cell_typ, netlist.wire_in, netlist.wire_out, netlist.cell_in, netlist.cell_out, netlist.inputs)

	# store and restore reordered netlist
	#netlist_parser.reorder_store(instance_r, cell_typ, wire_in_r, wire_out, cell_in_r, cell_out, outdir)
	#instance_r, cell_typ, wire_in_r, wire_out, cell_in_r, cell_out = netlist_parser.reorder_restore(netlist_r_text)

	print 'netlist tree covering...'
	
	# tree corvering
	global trees
	global trees_in
	global root_instances
	trees, trees_in, root_instances = netlist_parser.tree_covering(wire_in_r, wire_out, instance_r, primary_inputs, netlist.outputs)

	# generate input vectors
	input_vec = signal_probability.generate(len(primary_inputs), vec_no, seed)
	
	# read sp expression file
	sp_content = general.read_txt(sp_file)

	# parse sp expression
	global sp_typ
	global sp_out
	global sp_expression
	sp_typ, sp_out, sp_expression = signal_probability.sp_parse(sp_content)

	# work out sp for each node
	global node_order
	global sp_order
	node_order, sp_order = signal_probability.node_sp(primary_inputs, input_vec, cell_typ, wire_in_r, wire_out, cell_in_r, sp_typ, sp_out, sp_expression)
	
	# read stacking effect definition file
	stacking_content = general.read_txt(stacking_file)

	# parse stacking effect definition expression
	global stacking_typ_u
	global stacking_input_u
	global expression_u
	global stacking_typ_d
	global stacking_input_d
	global expression_d
	stacking_typ_u, stacking_input_u, expression_u, stacking_typ_d, stacking_input_d, expression_d = signal_probability.stacking_parse(stacking_content)

	# reoder cell types, input pins and instance names according to conenctions
	global cell_typ_order
	global cell_in_order
	global instance_order
	global node_order2
	cell_typ_order, cell_in_order, instance_order, node_order2 = signal_probability.others_reorder(node_order, cell_typ, cell_in_r, wire_in_r, instance_r)
	
	# sp mean and standard deviation
	print '\nInput vector done: calculating signal probabilities mean & deviation'
	global sp_mean
	sp_mean, sp_std = signal_probability.sp_mean_std(sp_order)

	# parse logic file
	logic_content = general.read_txt(logic_file)
	global logic_cell_typ
	global logic_expression
	global product_number
	global term_per_product
	global cell_in
	global in_occurrence
	global in_term_products
	logic_cell_typ, logic_expression, product_number, term_per_product, cell_in, in_occurrence, in_term_products = tree_matching.logic_parser(logic_content)

# *******************************************************************************

# *******************************************************************************
def evaluation():
	
	# run Design Compiler with tcl file in output directory
	print 'running Design Compiler in background ...'
	general.run_dc(outdir, tcl)

	print 'parsing area/timing/power/sdf report ...'

	# read DC report
	rise_timing_content = general.read_txt(outdir+rise_timing_rpt)	# timing
	fall_timing_content = general.read_txt(outdir+fall_timing_rpt)
	timing_content1 = rise_timing_content + fall_timing_content
	timing_content = [x.replace('<-', '') for x in timing_content1]
	sdf_content = general.read_txt(outdir+sdf_file)		# standard delay format
	area_content = general.read_txt(outdir+area_rpt)	# area

	# parse DC report
	global timing
	global area
	timing = timing_parser.analyze(timing_content)
	sdf = sdf_parser.analyze(sdf_content)
	area = area_parser.analyze(area_content)

	# work out cell input to input delay
	global new_delay
	global port_in
	new_delay, port_in = sdf_parser.in2in_delay(timing.instance, timing.delay, sdf.wire_in, sdf.wire_out, sdf.delay, sdf.wire_out_port)

	# stress duty cycles for all transistors of the critical paths
	global nbti_path
	nbti_path = signal_probability.find_alpha(port_in, instance_order, cell_in_order, cell_typ_order, sp_order, stacking_typ_u, stacking_input_u, expression_u, stacking_typ_d, stacking_input_d, expression_d, node_order, node_order2)

	# convert port name to wire name
	global wire_in
	wire_in = ageing.port2wire(port_in, instance_r, cell_in_r, wire_in_r)

	# applying ageing model
	global delta_mean
	global path_delay
	global delta
	delta_mean, path_delay, delta = ageing.iteration(vec_no, nbti_path, wire_in, new_delay, timing.transition, timing.delay)

	# percentage degradation
	global delta_percent
	delta_percent = [x/y*100 for x,y in zip(delta_mean, path_delay)]

	# produce critical instance list
	global re_mapping_list
	global cp_index
	re_mapping_list, cp_index = timing_parser.instance_list(timing, netlist, wire_in, delta_mean, timing_cons * 0.8)

	# record degradation for each input vector
	max_index = delta_mean.index(max(delta_mean))
	delta_record = open(outdir+delta_log_path, 'w')
	if delta_log_mode == 'latex':
		for each in delta:
			delta_record.write('%s\n' %each[max_index])
	elif delta_log_mode == 'matlab':
		for i, each in enumerate(delta):
			if i == 0:
				delta_record.write('[%s' %each[max_index])
			else:
				delta_record.write(',%s' %each[max_index])
		delta_record.write('];\n')

# *******************************************************************************


# *******************************************************************************
# optimization (technology remapping)
def optimization():

	new_out_nodes = []
	new_out_sps = []

	tree_structure = []

	# filter out non-critical trees
	critical_tree_index, critical_tree_instances = tree_matching.tree_filter(trees, re_mapping_list)

	# produce tree verilog netlist
	print '\nproducing critical tree verilog netlist ...'
	tree_area = []
	for index in critical_tree_index:
		
		tree_under_opt = trees[index]
		tree_in = trees_in[index]
		root_instance = root_instances[index]
		area = tree_matching.tree_netlist(min_delay_tree_netlist, tree_under_opt, tree_in, root_instance, netlist, index, lib)
		tree_matching.tree_netlist(min_deg_tree_netlist, tree_under_opt, tree_in, root_instance, netlist, index, lib)
		tree_area.append(area)
	
	# iteratively optimizate each tree
	print 'Forward pass:'
	for cnt, index in enumerate(critical_tree_index):

		tree_under_opt = trees[index]
		tree_in = trees_in[index]
		root_instance = root_instances[index]
		critical_instances = critical_tree_instances[cnt]

		# compute tree stress
		tree_paths, tree_nodes, tree_ports, tree_sps, tree_stress = tree_matching.tree_stress_computation(tree_under_opt, timing, cp_index, wire_in, port_in, sp_mean, node_order, nbti_path)
		
		# divide tree into subtree
		critical_instances = critical_tree_instances[cnt]
		non_critical_instances = [instance for instance in tree_under_opt if instance not in critical_instances]
		subtrees = tree_matching.divide_subtree(tree_under_opt, critical_instances, netlist)

		# optimize each subtree
		subtree_structure = []
		for subtree_i, subtree in enumerate(subtrees):
			
			# capture subtree info
			tree_netlist = general.read_txt(min_deg_tree_netlist)
			subtree_line_index, root_typ, root_port, root_wire, leave_typ, leave_port, leave_wire, ct_instances, ct_nodes = tree_matching.subtree_info(subtree, tree_netlist, timing, wire_in, cp_index)

			# figure out subtree logic
			subtree_logic = tree_matching.logic_analysis(subtree, root_typ, root_wire, logic_cell_typ, logic_expression, leave_typ, leave_wire)
			
			# restructure subtree logic
			optimal_subtree_logic = tree_matching.logic_restructuring(subtree, subtree_logic, ct_instances, ct_nodes)

			# map the logic
			optimal_subtree, optimal_cell_typ, optimal_nodes, org_subtree, org_cell_typ, org_nodes = tree_matching.logic_mapping(optimal_subtree_logic, logic_cell_typ, product_number, term_per_product, cell_in, in_occurrence, subtree, root_typ, root_wire, in_term_products, leave_typ, leave_wire)

			# recompute stress
			final_structure, out_nodes, out_sps, out_nodes_irr, out_sps_irr = tree_matching.tree_stress_recomputation(tree_nodes, tree_stress, optimal_subtree, optimal_cell_typ, optimal_nodes, org_subtree, sp_mean, node_order, stacking_typ_u, stacking_input_u, expression_u, sp_typ, sp_out, sp_expression, new_out_nodes, new_out_sps)
			new_out_nodes.extend(out_nodes)
			new_out_sps.extend(out_sps)
			new_out_nodes.extend(out_nodes_irr)
			new_out_sps.extend(out_sps_irr)

			# update tree netlist
			if final_structure == 'org':
				final_subtree = org_subtree
				final_cell_typ = org_cell_typ
				final_nodes = org_nodes
			else:
				final_subtree = optimal_subtree
				final_cell_typ = optimal_cell_typ
				final_nodes = optimal_nodes
			tree_matching.tree_netlist_update(tree_netlist, subtree_line_index, final_subtree, final_cell_typ, final_nodes, lib, min_deg_tree_netlist, critical_instances, netlist)
			
			# record subtree transformations
			subtree_structure.append(final_structure)

		# record tree transformations
		if len(subtree_structure) == 0 or 'new' not in subtree_structure:
			tree_structure.append('org')
		else:
			tree_structure.append('new')

		print '--------------------------------------------------------------------------'  
		# if structure does not change, restore the tree circuit
		if len(subtree_structure) == 0 or 'new' not in subtree_structure:
			tree_matching.tree_restore(min_deg_tree_netlist, min_delay_tree_netlist, index)
			print 'remapping tree %d/%d: no transformation has been made.' %((cnt+1), len(critical_tree_index))
			
		# if structure changed, resynthesize the tree circuit
		else:
			print 'remapping tree %d/%d: done.' %((cnt+1), len(critical_tree_index))

			if min_area_enable == False:
				all_delay_cons = []
				all_area_distance = []
			
				# gate count on optimal tree
				number_of_gate1 = tree_matching.gate_count(min_deg_tree_netlist, index)
				
				# maximal circuit delay
				max_delay = max(path_delay)
			
				# update tree synthesis scripts
				upper_lim = max_delay
				lower_lim = 0
				area_cons = tree_area[cnt]
				delay_cons = (upper_lim + lower_lim)/2
				first_time = True
				delay_line_index = 0
				new_delay_line_index = tree_matching.tree_tcl_update(tree_tcl, area_cons, first_time, delay_line_index, delay_cons, index, non_critical_instances)
				delay_line_index = new_delay_line_index
				first_time = False

				# resynthesize the tree
				print 'remapping tree %d/%d: delay constraint = %fns.' %((cnt+1), len(critical_tree_index), delay_cons)
				general.run_dc(outdir, tree_tcl)

				# check actual area
				area_content = general.read_txt(outdir+'/tree_area.rpt')
				area = area_parser.analyze(area_content)

				# gate count on resynthesized tree
				number_of_gate2 = tree_matching.gate_count('tree_temp.v', index)
				if number_of_gate2 == number_of_gate1:
					all_delay_cons.append(delay_cons)
					all_area_distance.append(abs(area-area_cons))
			
				trial_cnt = 1
				while (area > area_cons + 5 or area < area_cons - 5) and trial_cnt < max_trial:
					
					trial_cnt = trial_cnt + 1

					if area > area_cons or number_of_gate2 < number_of_gate1:
						lower_lim = delay_cons					
					elif area < area_cons:
						upper_lim = delay_cons
					delay_cons = (upper_lim + lower_lim)/2

					# update tree synthesis scripts
					tree_matching.tree_tcl_update(tree_tcl, area_cons, first_time, delay_line_index, delay_cons, index, non_critical_instances)

					# resynthesize the tree
					print 'remapping tree %d/%d: delay constraint = %fns.' %((cnt+1), len(critical_tree_index), delay_cons)
					general.run_dc(outdir, tree_tcl)

					# check actual area
					area_content = general.read_txt(outdir+'/tree_area.rpt')
					area = area_parser.analyze(area_content)

					# gate count on resynthesized tree
					number_of_gate2 = tree_matching.gate_count('tree_temp.v', index)
					if number_of_gate2 == number_of_gate1:
						all_delay_cons.append(delay_cons)
						all_area_distance.append(abs(area-area_cons))

				# if exceed the maximal number of trial, find the best
				if trial_cnt == max_trial:
					min_area_distance_index = all_area_distance.index(min(all_area_distance))
					delay_cons = all_delay_cons[min_area_distance_index]

					# update tree synthesis scripts
					tree_matching.tree_tcl_update(tree_tcl, area_cons, first_time, delay_line_index, delay_cons, index, non_critical_instances)

					# resynthesize the tree
					print 'remapping tree %d/%d: delay constraint = %fns.' %((cnt+1), len(critical_tree_index), delay_cons)
					general.run_dc(outdir, tree_tcl)

					# check actual area
					area_content = general.read_txt(outdir+'/tree_area.rpt')
					area = area_parser.analyze(area_content)
			
				print 'remapping tree %d/%d: final area = %fum^2 -> %fum^2.' %((cnt+1), len(critical_tree_index), area_cons, area)
			
				# change mininal degradation tree netlist
				tree_matching.tree_restore(min_deg_tree_netlist, 'tree_temp.v', index)

	# backward check the arrival time
	print 'Backward pass:'
	for cnt, index in enumerate(critical_tree_index):

		print '--------------------------------------------------------------------------' 

		# check tree structure
		structure = tree_structure[cnt]
		if structure == 'org':
			print 'tree %d/%d: no transformation has been made.' %((cnt+1), len(critical_tree_index))

		elif structure == 'new':

			tree_under_opt = trees[index]
			root_instance = root_instances[index]
			tree_instance = 'T' + str(index)
			
			# replace the tree in original netlist with optimized one
			backup_netlist = netlist_rename.split('.')[0] + '_backup.v'
			tree_matching.org_netlist_replace(outdir+netlist_rename, outdir+backup_netlist, min_deg_tree_netlist, index, tree_under_opt)

			#root node
			root_index = netlist.instance_backup.index(root_instance)
			tree_out = netlist.wire_out_backup[root_index]
			if '[' in tree_out[0]:
				root_node = '\\'+tree_out[0]
			else:
				root_node = tree_out[0]

			# modify backward pass synthesis scripts
			tree_matching.backward_tcl_update(backward_tcl, [netlist_rename.replace('/', ''), min_deg_tree_netlist], tree_instance, root_node)

			# run DC
			print 'running Design Compiler in background ...'
			general.run_dc(outdir, backward_tcl)

			# check arrival time
			arrival_time, req_time = tree_matching.arrival_time_checking(timing_cons + looser_timing, tree_instance, root_node)
			
			print 'tree %d/%d: required arrival at %s/%s (ns); actual arrival at %s/%s (ns).' %((cnt+1), len(critical_tree_index), req_time[0], req_time[1], arrival_time[0], arrival_time[1])

			time_difference = [float(req_time[i]) - float(arrival_time[i]) for i in range(0, len(req_time))]
			if min(time_difference) > 0:
				print 'tree %d/%d: transform tree structure into optimal one for ageing.' %((cnt+1), len(critical_tree_index))
			else:
				print 'tree %d/%d: restore tree structure into original one.' %((cnt+1), len(critical_tree_index))
				shutil.copy(outdir+backup_netlist, outdir+netlist_rename)
	
				
# *******************************************************************************


# ********************************************************************************
# modify timing report content with ageing degradation
def modify_timing_rpt():

	rise_rpt = general.read_txt(outdir + rise_timing_rpt)
	fall_rpt = general.read_txt(outdir + fall_timing_rpt)
	rpt = rise_rpt + fall_rpt
	new_rpt = rpt

	# find path sp mean for each cell
	path_sp = ageing.find_sp(node_order, sp_mean, wire_in, 0)
	
	# for each path
	for index, value in enumerate(timing.start):
		
		# add column name
		new_rpt[int(value)-2] = '%s\tAlpha\tDegradation(10yrs)\n' %rpt[int(value)-2].rstrip('\n')

		# add cell stress duty cycle and degradation
		cnt = 0
		for i in range(int(value)+2, int(timing.end[index])-1):
			if timing.transition[index][cnt+1] == 'r':
				cell_delta = ageing.ageing(float(new_delay[index][cnt]), float(nbti_path[index][cnt][0]))
				new_rpt[i] = '%s\t%f\t%f\n' %(rpt[i].rstrip('\n'), nbti_path[index][cnt][0], cell_delta)
			elif timing.transition[index][cnt+1] == 'f':
				cell_delta = 0.1 * ageing.ageing(float(new_delay[index][cnt]), 1-float(nbti_path[index][cnt][0]))
				#new_rpt[i] = '%s\t%f\n' %(rpt[i].rstrip('\n'), path_sp[index][cnt])
				new_rpt[i] = '%s\t%f\t%f\n' %(rpt[i].rstrip('\n'), 0.1 * path_sp[index][cnt], cell_delta)
				#new_rpt[i] = '%s\t%f\t%f\n' %(rpt[i].rstrip('\n'), 0, 0)

			cnt = cnt + 1
		
		# add overall degradation
		new_rpt[int(timing.end[index])] = '%s\t\t+%f(+%f%%)\n' %(rpt[int(timing.end[index])].rstrip('\n'), delta_mean[index], delta_percent[index])
	
	# write post-ageing timing report
	report = open(outdir + new_timing_rpt, 'a')
	for item in new_rpt:
		report.write(item)

	print '\nPost-ageing timing report complete!'
	
	path_delay_10yrs = [float(x)+float(y) for x, y in zip(path_delay, delta_mean)]
	print 'Design area: %s, max freq: %s ns, max freq(10 yrs): %s ns, max degradation(10 yrs): %s ns' %(area, max(path_delay), max(path_delay_10yrs), max(delta_mean))

# ********************************************************************************


# ********************************************************************************
# produce post-ageing timing report
def main():

	prepare()
	evaluation()

	start_time = time.time()

	if remap_enable == True:
		optimization()
	modify_timing_rpt()

	print 'Optimization takes %s seconds' %(time.time()-start_time)

# ********************************************************************************

main()
