import general, re, lib_parser

# *******************************************************************************
# analyze verilog structural netlist (content = content.readlines())
class analyze:
	
	def __init__(self, content, cell_info):

		# primary inputs
		in_def = self.input_line(content)
		self.inputs = self.input_parser(in_def)

		# primary outputs
		out_def = self.output_line(content)
		self.outputs = self.output_parser(out_def)

		# instance
		self.cell_def = self.cell_line(content, cell_info.cell_name)
		self.instance, self.cell_typ, self.wire_in, self.wire_out, self.cell_in, self.cell_out, self.cell_size = self.instance_parser(self.cell_def, cell_info)
		self.instance_backup, self.cell_typ_backup, self.wire_in_backup, self.wire_out_backup, _, _, _ = self.instance_parser(self.cell_def, cell_info)

	# find input lines
	def input_line(self, content):
		in_def = []
		input_line_end = True

		for line in content:
			if 'input' in line or input_line_end == False:
				in_def.append(line)

				if ';' in line:
					input_line_end = True
				else:
					input_line_end = False
		
		# find lines with input information
		return in_def

	# find output lines
	def output_line(self, content):
		out_def = []
		output_line_end = True

		for line in content:
			if 'output' in line or output_line_end == False:
				out_def.append(line)

				if ';' in line:
					output_line_end = True
				else:
					output_line_end = False
		
		# find lines with output information
		return out_def


	# find cell/instance blocks
	def cell_line(self, content, cell_name):

		lines = []
		cell_re = re.compile("|".join(cell_name))
		
		for index, line in enumerate(content):

			# if cell name in library exists in the line
			if cell_re.search(line):
				
				# if ; in the line
				if ';' in line:
					lines.append(line)

				# if ; not in the line
				else:
					cnt = 1
					joint_line = line.replace('\n', '')
					while ';' not in content[index+cnt]:
						cnt = cnt + 1
						joint_line = joint_line + content[index+cnt].replace('\n', '')
					joint_line = joint_line + content[index+cnt].replace('\n', '')
					lines.append(joint_line)
		
		return lines
	
	
	# parse input names
	def input_parser(self, in_def):
	
		inputs = []
		
		# define format
		format = re.compile('.*[.*:.*].*')	
	
		for line in in_def:
	
			# if format = input [x:y] A
			if format.match(line) is not None:
				digit_1 = line[line.find("[")+1:line.find(":")]
				digit_2 = line[line.find(":")+1:line.find("]")]
				name = line[line.find("]")+1:line.find(";")].replace(" ", "")
				for i in range(int(digit_2), int(digit_1)+1):
					inputs.append(name+"[%d]" % i)
			
			# if format = input A,B,...
			else:
				new_line = line.replace("input", "").replace(" ", "").replace(";", "").replace("\n", "").replace("\\", "")
				inputs.extend(new_line.split(","))
		
		inputs = filter(None, inputs)
		return inputs

	# parse input names
	def output_parser(self, out_def):
	
		outputs = []
		
		# define format
		format = re.compile('.*[.*:.*].*')	
	
		for line in out_def:
	
			# if format = output [x:y] A
			if format.match(line) is not None:
				digit_1 = line[line.find("[")+1:line.find(":")]
				digit_2 = line[line.find(":")+1:line.find("]")]
				name = line[line.find("]")+1:line.find(";")].replace(" ", "")
				for i in range(int(digit_2), int(digit_1)+1):
					outputs.append(name+"[%d]" % i)
			
			# if format = output A,B,...
			else:
				new_line = line.replace("output", "").replace(" ", "").replace(";", "").replace("\n", "").replace("\\", "")
				outputs.extend(new_line.split(","))
		
		outputs = filter(None, outputs)
		return outputs
	
	
	# parse instance/cell information
	def instance_parser(self, cell_def, cell_info):

		instance = []	# instance name
		cell_typ = []	# cell type
		wire_in = []	# input wire
		wire_out = []	# output wire
		cell_in = []	# cell input 
		cell_out = []	# cell output
		cell_size = []	# cell size

		# for each instance in netlist
		for i, line in enumerate(cell_def):
			
			# instance list
			instance.append(line.split()[1])
			
			# cell list
			cell_name = line.split()[0]
			size = re.findall('\d+', cell_name)
			cnt = len(size[-1])
			cell_typ.append(cell_name[:-cnt])
			cell_size.append(size[-1])

			port_list = []	# port name
			wire_list = []	# wire name
			cell_in.append([])
			cell_out.append([])
			wire_in.append([])
			wire_out.append([])
			port_def =  line.split('(',1)[1].split(',')
			
			for value in range(0, len(port_def)):
			
				# port list
				port_name = port_def[value][port_def[value].find(".")+1:port_def[value].find("(")]
				port_list.append(port_name)
			
				# wire list
				wire_name = port_def[value][port_def[value].find("(")+1:port_def[value].find(")")]
				wire_list.append(wire_name)
			
			_, port_in, port_out = lib_parser.find_cell(cell_name[:-cnt], cell_info.cell_typ, cell_info.cell_size, cell_info.cell_name, cell_info.pin_in, cell_info.pin_out)
			
			for port_index, port in enumerate(port_list):
				
				# for input
				if port in port_in:
					#wire_in[i].append(wire_list[port_index].replace(" ", "").replace("\\", ""))
					cell_in[i].append(port_list[port_index].replace(" ", "").replace("\\", ""))
					wire_in[i].append(wire_list[port_index].replace(" ", ""))

				# for output
				if port in port_out:
					#wire_out[i].append(wire_list[port_index].replace(" ", "").replace("\\", ""))
					cell_out[i].append(port_list[port_index].replace(" ", "").replace("\\", ""))					
					wire_out[i].append(wire_list[port_index].replace(" ", ""))
			
		return instance, cell_typ, wire_in, wire_out, cell_in, cell_out, cell_size

# *******************************************************************************


# *******************************************************************************
# reorder netlist according to connections
def reorder(instance, cell_typ, wire_in, wire_out, cell_in, cell_out, primary_input):
	
	# reordered information
	instance_r = []
	cell_typ_r = []
	wire_in_r = []
	wire_out_r = []
	cell_in_r = []
	cell_out_r = []
	
	# nodes connecting to known nodes
	known_node = primary_input

	cnt = 0
	
	# if there is information not analyzed yet
	while wire_in:

		print len(wire_in)

		for i, node in enumerate(wire_in):

			# node belonging to known nodes
			if set(node) <= set(known_node):
		
				# update known nodes
				known_node = known_node + wire_out[i]
	
				# update reordered information
				instance_r.append(instance[i])
				cell_typ_r.append(cell_typ[i])
				cell_in_r.append(cell_in[i])
				cell_out_r.append(cell_out[i])
				wire_in_r.append(wire_in[i])
				wire_out_r.append(wire_out[i])

				# delete information that is already analyzed
				del instance[i]
				del cell_typ[i]
				del cell_in[i]
				del cell_out[i]
				del wire_in[i]
				del wire_out[i]

	return instance_r, cell_typ_r, wire_in_r, wire_out_r, cell_in_r, cell_out_r

# *******************************************************************************


# *******************************************************************************
# remove '\' from netlist
def virgule_removal(netlist_content, path):
	
	# update netlist
	netlist_path = open(path, 'w')
	for line in netlist_content:
		
		# find instance name
		if len(line.split()) >= 2:
			instance_name = line.split()[1]
			new_instance_name = instance_name.replace('\\','').replace('/','')

			# check if instance name contains virgule
			found = re.search('\\T[0-9]+?/U[0-9]+?', instance_name)
			if found:
				new_line = line.replace(instance_name, new_instance_name)
			else:
				new_line = line

		else:
			new_line = line
	
		netlist_path.write(new_line)
	netlist_path.close()

# *******************************************************************************


# *******************************************************************************
# divide the circuit into trees
def tree_covering(wire_in, wire_out, instance, primary_in, primary_out):

	# all fanout>2 nodes
	fanout_nodes = []
	flatten_wire_in = [item for sublist in wire_in for item in sublist]
	for node in flatten_wire_in:
		
		if flatten_wire_in.count(node) > 1:
			fanout_nodes.append(node)
	
	# root nodes
	roots = list(set(fanout_nodes+primary_out))

	# top leaves
	leaves = list(set(fanout_nodes+primary_in))

	# tree covering
	tree = []

	# tree input
	tree_in = []

	# tree output
	tree_out = []

	instance_covered = [False] * len(instance)
	flatten_wire_out = [item for sublist in wire_out for item in sublist]
	
	cnt = 0
	for index, covered_or_not in enumerate(reversed(instance_covered)):
		true_index = len(instance) - index - 1
		
		if covered_or_not == False:
			instance_covered[true_index] = True
			tree.append([])
			tree[cnt].append(instance[true_index])
			tree_in.append([])

			# loop until all input nodes are top leaves
			input_nodes = wire_in[true_index]
			while (set(input_nodes) < set(leaves)) == False:

				input_nodes_update = []
				
				for input_node in input_nodes:			
	
					if input_node not in leaves:
						
						# put instance name into current tree
						input_index = flatten_wire_out.index(input_node)
						tree[cnt].append(instance[input_index])
						instance_covered[input_index] = True

						# update input nodes
						input_nodes_update.extend([item for item in wire_in[input_index]])
					
					else:
						input_nodes_update.append(input_node)
							
				input_nodes = list(set(input_nodes_update))
		
			# tree input
			tree_in[cnt].append(input_nodes)	

			cnt = cnt + 1

	# root instances
	root_instances = []
	for each_tree in tree:
		
		root_instances.append(each_tree[0])

	return tree, tree_in, root_instances

# *******************************************************************************


# *******************************************************************************
# store reordered netlist
def reorder_store(instance_r, cell_typ, wire_in_r, wire_out, cell_in_r, cell_out, outdir):
	netlist_r_text = open(outdir+'/netlist_reorder.txt', 'w')
	for i, each in enumerate(instance_r):
		netlist_r_text.write('%s\t%s\t%s\t%s\t%s\t%s\n' %(each, cell_typ[i], wire_in_r[i], wire_out[i], cell_in_r[i], cell_out[i]))
			
# *******************************************************************************


# *******************************************************************************
# restore reordered netlist
def reorder_restore(netlist_r_text):
	netlist_r_content = general.read_txt(netlist_r_text)
	instance_r = []
	cell_typ = []
	wire_in_r = []
	wire_out_r = []
	cell_in_r = []
	cell_out_r = []
	for i,line in enumerate(netlist_r_content):

		# instance and cell type
		instance_r.append(line.split()[0])
		cell_typ.append(line.split()[1])

		# wire_in and cell_in
		wire_in = line.split('\t')[2].replace('[','').replace(']','').replace('\'','')
		wire_in_r.append([])
		for each in wire_in.split():
			wire_in_r[i].append(each.replace(',',''))

		cell_in = line.split('\t')[4].replace('[','').replace(']','').replace('\'','')
		cell_in_r.append([])
		for each in cell_in.split():
			cell_in_r[i].append(each.replace(',',''))

		# wire_out and cell_out
		wire_out = line.split('\t')[3].replace('[','').replace(']','').replace('\'','')
		wire_out_r.append([])
		for each in wire_out.split():
			wire_out_r[i].append(each.replace(',',''))

		cell_out = line.split('\t')[5].replace('[','').replace(']','').replace('\'','')
		cell_out_r.append([])
		for each in cell_out.split():
			cell_out_r[i].append(each.replace(',',''))

	return instance_r, cell_typ, wire_in_r, wire_out_r, cell_in_r, cell_out_r
			
# *******************************************************************************


# *******************************************************************************
# test program
def test():
	netlist_path = '../dc_syn/test.v'
	netlist = general.read_txt(netlist_path)
	
	library = '../verilog/CORE65LPSVT_bc_1.05V_105C.lib'
	lib = general.read_txt(library)
	lib_info = lib_parser.analyze(lib)
	
	netlist_info = analyze(netlist, lib_info)
	print netlist_info.inputs
	print netlist_info.wire_in
	
	instance_r, cell_typ_r, wire_in_r, wire_out_r, cell_in_r, cell_out_r = reorder(netlist_info.instance, netlist_info.cell_typ, netlist_info.wire_in, netlist_info.wire_out, netlist_info.cell_in, netlist_info.cell_out, netlist_info.inputs)
	print instance_r
	print cell_in_r
	print wire_in_r

#test()

# *******************************************************************************		
