import general, os, math, random

# ********************************************************************************
# update netlist (upsize)
def update_netlist(netlist_old, netlist_new, instance, cell, opt_cell, max_size):
	
	# read netlist
	content = general.read_txt(netlist_old)

	# modify selected cell size
	if max_size == False:
		for i, line in enumerate(content):
			if instance in line and cell in line:
				content[i] = line.replace(cell, opt_cell)

	# update netlist
	netlist = open(netlist_new, 'w')
	for item in content:
		netlist.write(item)
	netlist.close()

# *******************************************************************************


# ********************************************************************************
# sensitivity analysis (delay or degradation/area) by upsizing
def sensitivity(x_decrease_record, y_decrease_record, area_increase_record, area_record, max_area, area0, min_delay, delay0, seed, path_delay_record):
	
	SA_ena = False
	random.seed(seed)	

	sens = []

	optimization_end = False
	
	# (delay or degradation)/area sensitivity
	#for x, y in zip(x_decrease_record, area_increase_record):
	#	if x == 0:
	#	 	sens.append(0)
	#	elif y == 0:
	#	 	sens.append(x*float('inf'))
	#	else:
	#	 	sens.append(x/y)

	# new sensitivity
	for index, x in enumerate(x_decrease_record):
		
		area_percent = (max_area - area_record[index]) / area_record[index]
		delay_percent = (path_delay_record[index]- min_delay) / path_delay_record[index]
		
		if max_area > area_record[index]:
			value = x + (3.2/10) * (delay_percent/area_percent) * y_decrease_record[index]
		else:
			value = y_decrease_record[index]

		if area_increase_record[index] == 0:
			sens.append(-float('inf'))
		elif value == 0:
			sens.append(0)
		else:
			sens.append(value/area_increase_record[index])

	# sort sensitivity
	sens_x = zip(sens, x_decrease_record)

	sens_x.sort(key=lambda sl: (-sl[0],-sl[1]))

	# best solution index
	cnt = 0
	best_index = zip(sens, x_decrease_record).index(sens_x[cnt])

	# if y decrease negative (i.e., delay increases)
	while y_decrease_record[best_index] <= 0 and optimization_end == False and SA_ena == True:
			
		# simulated annealing
		T = (max_area - area_record[best_index])/(max_area - area0)	# temperature
		delta_D = -0.25 + y_decrease_record[best_index] * 0.4 / 0.001	# change of distance
		if T > 0:
			p = math.exp(delta_D/T)	# acceptance probability
		else:
			p = 0

		print p
			
		# if beyond acceptance probability
		if random.random() > p:
			print 'reject potential decision of trial %d' %best_index
			cnt = cnt + 1

			# if not exceeding path length
			if cnt < len(sens):
				best_index = zip(sens, x_decrease_record).index(sens_x[cnt])
			else:
				optimization_end = True

		else:
			print 'accept potential decision of trial %d' %best_index
			break

	return best_index, optimization_end, sens

# *******************************************************************************
