import general, re

# ******************************************************************************
# analyze DC timing report (content = content.readlines())
class analyze:

	def __init__(self, content):
		block, self.start, self.end = self.path_block(content)
		self.instance, self.cell_name, self.delay, self.transition, self.port_name = self.cell_info(block)
		self.cell_typ, self.cell_size = self.split_typ_size(self.cell_name)
	

	# find path blocks
	def path_block(self, content):
		# find start/end lines of path block
		start = []
		end = []
		for number, line in enumerate(content):
			if 'input external delay' in line:
				start.append(number)
			if 'data arrival time' in line and '-' not in line:
				end.append(number)
		# find path blocks
		block = []
		if len(start) != len(end):
			print 'Error when extract blocks'
		else:
			for index, number in enumerate(start):
				block.append([])
				block[index].append(content[number+1:end[index]-1])
		return block, start, end

	
	# cell information
	def cell_info(self, block):

		instance = []	# instance name
		cell_name = []	# cell name
		delay = []	# cell delay
		transition = []	# type of transition
		port_name = []	# output name

		# for each path of the blocks
		for index, path in enumerate(block):
			instance.append([])
			cell_name.append([])
			delay.append([])
			transition.append([])
			port_name.append([])

			# for each cell of the path
			for cell in path[0]:
				cell.replace('<-', '')

				instance[index].append(cell.split(None,2)[0].split('/',1)[0].replace(' ',''))

				#instance_line = cell.split(None,2)[0].replace(' ','').split('/')
				#if len(instance_line) == 1:
				#	instance[index].append("/".join(instance_line))
				#else:
				#	instance[index].append("/".join(instance_line[:-1]))

				cell_name[index].append(cell[cell.find('(')+1:cell.find(')')])
				delay[index].append(cell.split(None,5)[2])
				transition[index].append(cell.split(None,5)[4])
				port_name[index].append(cell.split(None,2)[0])

		return instance, cell_name, delay, transition, port_name
	
	# split cell type and size
	def split_typ_size(self, cell_name):
		gate_size = []
		gate_typ = []

		# for each path	
		for i, path in enumerate(cell_name):
		
			gate_size.append([])
			gate_size[i].append([])

			gate_typ.append([])
			gate_typ[i].append([])
		
			# for each cell except the input 
			for cell in path[1:]:
			
				size = re.findall('\d+', cell)
			
				cnt = len(size[-1])
				gate_size[i].append(int(size[-1]))
				gate_typ[i].append(cell[:-cnt])
	
		return gate_typ, gate_size
	
# ******************************************************************************


# ******************************************************************************
# generate list of critical instances
def instance_list(timing, netlist, wire_in, delta_mean, threshold):

	re_mapping_list = []
	#critical_nodes = []
	critical_path_index = []	

	# for each path
	for index, path in enumerate(timing.delay):
		path_delay =  sum([float(each) for each in path])
		
		if path_delay >= threshold:

			# add instance of bti critical paths into the list
			re_mapping_list.extend(timing.instance[index][1:])

			# add path index into critical path index
			critical_path_index.append(index)

			# add node of bti critical paths into the list
			#critical_nodes.extend(wire_in[index])

	re_mapping_list = list(set(re_mapping_list))

	# critical nodes for each critical instances
	#ct_instance_nodes = []
	#critical_nodes = list(set(critical_nodes))
	#for instance in re_mapping_list:
	#	instance_index = netlist.instance_backup.index(instance)
	#	instance_wire_in = netlist.wire_in_backup[instance_index]
	#	common_nodes = list(set(critical_nodes).intersection(instance_wire_in))
	#	ct_instance_nodes.append(common_nodes)

	return re_mapping_list, critical_path_index

# ******************************************************************************


# ******************************************************************************
# find all instance that can be re-mapped 
def find_instance_remap(instance_path, instance_transition, level_th_percent, root_instances, tree):

	# circuit logic level
	logic_level = len(max(instance_path, key=len)) - 1
	
	# path level threshod
	path_level_th = logic_level * level_th_percent

	# combine path instance with input transition
	for ele_instance, ele_transition in zip(instance_path, instance_transition):
		ele_instance.append(ele_transition[0])

	instance_path = [list(i) for i in set(map(tuple, instance_path))]
	
	# identify all instance that may be remapped according to level/transistion
	remap_list = []
	for path in instance_path:
		
		# check logic level
		if len(path) - 2 >= path_level_th:
		
			# input rising
			if path[-1] == 'r':
				remap_list.extend(path[0:-2:2])

			# input falling
			else:
				remap_list.extend(path[1:-2:2])
	
	remap_list = list(set(remap_list))

	print remap_list

	# filter out root instances
	for remap_instance in remap_list:
		
		if remap_instance.split('/')[0] in root_instances:
			remap_list.remove(remap_instance)
	
	print remap_list

# ******************************************************************************


# *******************************************************************************
# test program
def test():
	rpt_path = '../dc_syn/test_timing.rpt'

	rpt = general.read_txt(rpt_path)
	rpt_info = analyze(rpt)
	print rpt_info.cell_name
	print rpt_info.cell_typ
	print rpt_info.cell_size
	print rpt_info.delay

#test()

# *******************************************************************************
