import string, re, signal_probability, general, subprocess, area_parser, os, shutil

# *******************************************************************************
# filter out non-critical trees 
def tree_filter(trees, critical_instances):
	
	# critical tree index
	ct_index = []

	# critical instances for each tree
	ct_instances = []

	for index, tree in enumerate(trees):
		
		common_instances = []
		for instance in tree:
			if instance in critical_instances:
				common_instances.append(instance)
		
		if len(common_instances) >= 1 and len(tree) >= 2:
			
			# critical tree index
			ct_index.append(index)

			# critical instances for each tree
			ct_instances.append(common_instances)

	return ct_index, ct_instances

# *******************************************************************************


# *******************************************************************************
# produce verilog netlist for tree circuit
def tree_netlist(netlist_file, tree, tree_in, root_instance, org_netlist, count, lib):
	
	content = ''

	# input
	tree_in2 = []
	for inport in tree_in[0]:
		if '[' in inport:
			tree_in2.append('\\'+inport)
		else:
			tree_in2.append(inport)
	inports = '\t,'.join(tree_in2)

	#output
	root_index = org_netlist.instance_backup.index(root_instance)
	tree_out = org_netlist.wire_out_backup[root_index]
	if '[' in tree_out[0]:
		outport = '\\'+tree_out[0]
	else:
		outport = tree_out[0]

	# module definition	
	module_def = 'module\ttree%d\t(%s\t,%s\t);\n' %(count, inports, outport)
	content = content + module_def

	# ports definition
	ports_def = 'input\t%s\t;\noutput\t%s\t;\n' %(inports, outport)
	content = content + ports_def

	# cell definition
	area = 0
	for cell in tree:
		
		# cell index in original netlist
		index = org_netlist.instance_backup.index(cell)
		
		# cell definition in original netlist
		cell_def_org = org_netlist.cell_def[index].replace('\n','')

		content = content + cell_def_org + '\n'

		# work out tree area
		cell_name = cell_def_org.split()[0]
		area = area + lib.area[lib.cell_name.index(cell_name)]

	# endmodule
	content = content + 'endmodule' + '\n\n'

	# write tree verilog netlist
	netlist = open(netlist_file, 'a')
	netlist.write(content)

	return area

# *******************************************************************************


# *******************************************************************************
# divide tree into subtree
def divide_subtree(tree, critical_instances, org_netlist):

	# instance index
	instance_index = [org_netlist.instance_backup.index(instance) for instance in tree]

	# instance wire in/out
	wire_in = [org_netlist.wire_in_backup[index] for index in instance_index]
	wire_out = [org_netlist.wire_out_backup[index] for index in instance_index]
	
	subtrees = []
	
	for index, each in enumerate(critical_instances[1:]):

		cells = []

		# critical instance index
		ct_index = tree.index(each)

		# find the cell it drives
		outport = wire_out[ct_index][0]
		driving_or_not = [(outport in inports) for inports in wire_in]
		driving_cell = tree[driving_or_not.index(True)]
		cells.append(driving_cell)

		# find all driven cells
		cells.append(each)
		#driven_in = wire_in[driving_or_not.index(True)]
		#flatten_wire_out = [item for sublist in wire_out for item in sublist]
		#for each_driven_in in driven_in:
		#	if each_driven_in in flatten_wire_out:
		#		driven_cell = tree[flatten_wire_out.index(each_driven_in)]
		#		cells.append(driven_cell)
		
		if cells not in subtrees:
			subtrees.append(cells)


	subtrees = list(reversed(subtrees))

	return subtrees

# *******************************************************************************


# *******************************************************************************
# capture subtree info
def subtree_info(subtree, tree_netlist, timing, wire_in, cp_index):

	subtree_line_index = len(subtree)*[0]
	
	for index, line in enumerate(tree_netlist):

		# find all required cell definitions
		if len(line.split()) > 1 and line.split()[1] in subtree:
			subtree_index = subtree.index(line.split()[1])
			subtree_line_index[subtree_index] = index

	# capture root instance info
	root_line = tree_netlist[subtree_line_index[0]]

	# root cell type
	root_name = root_line.split()[0]
	root_size = re.findall('\d+', root_name)
	cnt = len(root_size[-1])
	root_typ = root_name[:-cnt]

	# root port/wire in/out
	root_port = []
	root_wire = []
	port_def =  root_line.split('(',1)[1].split(',')	
	for value in range(0, len(port_def)):
			
		# port list
		port_name = port_def[value][port_def[value].find(".")+1:port_def[value].find("(")]
		root_port.append(port_name)
	
		# wire list
		wire_name = port_def[value][port_def[value].find("(")+1:port_def[value].find(")")].replace(" ", "").replace("\\", "")
		root_wire.append(wire_name)

	# capture leaf instance info
	leave_typ = []
	leave_port = []
	leave_wire = []
	for line_index_index, line_index in enumerate(subtree_line_index[1:]):
		leaf_line = tree_netlist[line_index]

		# leaf cell typ
		leaf_name = leaf_line.split()[0]
		leaf_size = re.findall('\d+', leaf_name)
		cnt = len(leaf_size[-1])
		leave_typ.append(leaf_name[:-cnt])

		# leaf port/wire in/out
		leave_port.append([])
		leave_wire.append([])
		port_def = leaf_line.split('(',1)[1].split(',')
		
		for value in range(0, len(port_def)):
			
			# port list
			port_name = port_def[value][port_def[value].find(".")+1:port_def[value].find("(")]
			leave_port[line_index_index].append(port_name)

			# wire list
			wire_name = port_def[value][port_def[value].find("(")+1:port_def[value].find(")")].replace(" ", "").replace("\\", "")
			leave_wire[line_index_index].append(wire_name)

	# critical path instances/nodes
	all_cp_instances = [timing.instance[index] for index in cp_index]
	all_cp_nodes = [wire_in[index] for index in cp_index]

	critical_instances = []
	critical_nodes = []
	for path_index, path in enumerate(all_cp_instances):	
		for instance_index, instance in enumerate(path):

			# critical instances in subtree
			if instance in subtree[1:]:
				critical_instances.append(instance)

				# critical nodes in subtree
				critical_nodes.append(all_cp_nodes[path_index][instance_index-1])	

	return subtree_line_index, root_typ, root_port, root_wire, leave_typ, leave_port, leave_wire, critical_instances, critical_nodes	

# *******************************************************************************


# *******************************************************************************
# capture tree critical paths
def tree_cp(tree, timing, port_in, path_stress, wire_in, cp_index, root_instance):

	# critical path instances/ports/alpha/nodes/transitions
	cp_instances = [timing.instance[index] for index in cp_index]
	cp_ports = [port_in[index] for index in cp_index]
	cp_alpha = [path_stress[index] for index in cp_index]
	cp_nodes = [wire_in[index] for index in cp_index]
	cp_transitions = [timing.transition[index] for index in cp_index]

	tree_cp_instances = []
	tree_cp_ports = []
	tree_cp_alpha = []
	tree_cp_nodes = []
	tree_cp_transitions = []

	for index, path in enumerate(cp_instances):

		cnt = 0

		# find path related to tree
		if root_instance in path:
	
			root_location = path.index(root_instance)
			top_leaf_location = root_location

			# find top leaf locations
			while path[top_leaf_location] in tree:
				top_leaf_location = top_leaf_location - 1
			top_leaf_location = top_leaf_location + 1
	
			tree_cp_instances.append(path[top_leaf_location:root_location+1])
			tree_cp_ports.append(cp_ports[index][top_leaf_location-1:root_location])
			tree_cp_alpha.append(cp_alpha[index][top_leaf_location-1:root_location])
			tree_cp_nodes.append(cp_nodes[index][top_leaf_location-1:root_location])
			tree_cp_transitions.append(cp_transitions[index][top_leaf_location-1:root_location])

	return tree_cp_instances, tree_cp_nodes

# *******************************************************************************


# *******************************************************************************
# subtree logic analysis
def logic_analysis(subtree, root_typ, root_wire, logic_cell_typ, logic_expression, leave_typ, leave_wire):

	subtree_logic = []

	for i in range(0, len(subtree)):

		# root cell info
		if i == 0:
			wire_in = root_wire[:-1]
			wire_out = root_wire[-1]
			cell_typ = root_typ

		# leaf cells info (wrong, instance is not in original netlist, but in modified one)
		else:
			instance_index = i - 1
			wire_in = leave_wire[instance_index][:-1]
			wire_out = leave_wire[instance_index][-1]
			cell_typ = leave_typ[instance_index]

		# instance logic expression
		logic = logic_expression[logic_cell_typ.index(cell_typ)]

		# logic expression in terms of nodes
		logic = logic.replace('Zout', wire_out)
		for index, node in enumerate(wire_in):
			port_in = string.uppercase[index] + 'in'
			logic = logic.replace(port_in, node)
		subtree_logic.append(logic)

	return subtree_logic

# *******************************************************************************


# *******************************************************************************
# parse logic file
def logic_parser(logic_content):
	
	cell_typ = []
	cell_logic = []
	product_number = []
	cell_in = []
	in_occurrence = []
	term_per_product = []
	in_term_product = []

	# parse cell type and logic
	for line in logic_content:

		line_split = line.split()
		cell_typ.append(line_split[0])
		cell_logic.append(line_split[1])

	for index, logic in enumerate(cell_logic):

		sop = logic.split('=')[1]
		
		# parse number of product
		products = sop.split('|')
		logic_product_number = len(products)
		product_number.append(logic_product_number)

		# parse number of term per product
		term_per_product.append([len(product.split('&')) for product in products])

		# parse input pin
		all_pin_in = re.split('[| &]', sop)
		pin_in = list(set(all_pin_in))
		cell_in.append(pin_in)

		# input pin occurence
		in_term_product.append([])
		in_occurrence.append([])
		for i, pin in enumerate(sorted(pin_in)):
			each_occurrence = all_pin_in.count(pin)
			in_occurrence[index].append(each_occurrence)

			# number of term for the product contain a input pin
			in_term_product[index].append([])
			for product in products:
				if pin in product:
					number_of_term = len(product.split('&'))
					in_term_product[index][i].append(number_of_term)

	return cell_typ, cell_logic, product_number, term_per_product, cell_in, in_occurrence, in_term_product

# *******************************************************************************


# *******************************************************************************
# restructure logic
def logic_restructuring(subtree, subtree_logic, ct_instances, ct_nodes):

	optimal_subtree_logic = []

	out_nodes = []
	out_nodes_replace = []
	count = 0

	# restructure leaf logic
	for index, instance in enumerate(subtree[1:]):

		index = index + 1

		# check if instance is critical
		if instance in ct_instances:
			
			# extract sum of product terms
			sop = subtree_logic[index].split('=')[1]
			products = sop.split('|')

			# extract output
			out = subtree_logic[index].split('=')[0].replace('~', '')		

			# find critical nodes in each product
			product_ct_nodes = []
			for product in products:
				nodes = product.split('&')
				common_nodes = list(set(nodes).intersection(ct_nodes))
				product_ct_nodes.append(common_nodes)

			ct_nodes_comb = [list(i) for i in set(map(tuple, product_ct_nodes))]

			# check if the logic restructurable
			if len(ct_nodes_comb) > 1:

				out_nodes_replace.append([])
				out_nodes.append(out)

				# find products contain the same critical nodes
				cnt = 0
				for comb in ct_nodes_comb:
					products_comb = [products[i] for i, value in enumerate(product_ct_nodes) if value == comb]
					new_sop = '|'.join(products_comb)
					new_out = out + 'r' + str(cnt)
					out_nodes_replace[count].append(new_out)
					cnt = cnt + 1
					optimal_logic = '~' + new_out + '=' + new_sop
					optimal_subtree_logic.append(optimal_logic)

				count = count + 1
			
			else:
				optimal_subtree_logic.append(subtree_logic[index])

		else:
			optimal_subtree_logic.append(subtree_logic[index])

	# restructure root logic
	root_logic = subtree_logic[0]

	# replace nodes in root logic
	if len(out_nodes_replace) > 0:

		# extract sum of product terms
		sop = root_logic.split('=')[1]
		products = sop.split('|')

		for product in products:
			nodes = product.split('&')
			new_nodes = []
			for node in nodes:
				if node in out_nodes:
					node_index = out_nodes.index(node)
					new_nodes.extend(out_nodes_replace[node_index])
				else:
					new_nodes.append(node)
			new_product = '&'.join(new_nodes)
			root_logic = root_logic.replace(product, new_product)
	
	optimal_subtree_logic.insert(0, root_logic)

	return optimal_subtree_logic			

# *******************************************************************************


# *******************************************************************************
# logic mapping
def logic_mapping(subtree_logic, cell_typ, product_number, term_per_product, cell_in, in_occurrence, subtree, root_typ, root_nodes, in_term_product, leave_typ, leave_nodes):

	optimal_subtree = []
	optimal_cell_typ = []
	optimal_nodes = []
	match = True

	next_instance_index = 0
	pattern = re.compile('(^n[0-9]+r[0-9]+)', re.IGNORECASE)

	# find match for each logic
	for index, logic in enumerate(subtree_logic):

			optimal_nodes.append([])

			# output
			out = logic.split('=')[0].replace('~', '')				

			# work out number of product term
			sop = logic.split('=')[1]
			products = sop.split('|')
			logic_product_number = len(products)

			# parse number of term per product
			logic_term_per_product = [len(product.split('&')) for product in products]

			# work out input node occurrence
			logic_in_occurrence = []
			logic_in_term_product = []
			all_node_in = re.split('[| &]', sop)
			node_in = list(set(all_node_in))
			for node_index, node in enumerate(node_in):
				each_occurrence = all_node_in.count(node)
				logic_in_occurrence.append(each_occurrence)

				# number of term for the product contain a input pin
				logic_in_term_product.append([])
				for product in products:
					terms = product.split('&')
					if node in terms:
						number_of_term = len(terms)
						logic_in_term_product[node_index].append(number_of_term)

			# find matched cell
			matched_index = [i for i in range(0,len(cell_typ)) if product_number[i] == logic_product_number and sorted(in_occurrence[i]) == sorted(logic_in_occurrence) and sorted(term_per_product[i]) == sorted(logic_term_per_product)]
			
			# check if there is a match
			if len(matched_index) == 0:
				match = False
				break

			# if there is a match, update and find match for others	
			else:

				# find input/output nodes
				mapped_nodes = []
				cell_in_occurrence = in_occurrence[matched_index[0]]
				cell_in_term_product = in_term_product[matched_index[0]]
				for i in range(0,len(node_in)):
					port_in_occurrence = cell_in_occurrence[i]
					port_in_term_product = cell_in_term_product[i]
					for node_index, node in enumerate(node_in):
						if node not in mapped_nodes:
							if logic_in_occurrence[node_index] == port_in_occurrence and sorted(logic_in_term_product[node_index]) == sorted(port_in_term_product):
								mapped_nodes.append(node)
								optimal_nodes[index].append(node)
				optimal_nodes[index].append(out)
		
				# update instance name
				if index == len(subtree_logic) - 1:			
					if bool(pattern.match(out)) == False:
						new_instance = subtree[next_instance_index]
					else:
						sub_instance_no = out.split('r')[1]
						new_instance = subtree[next_instance_index] + 'r' + sub_instance_no
				else:
					next_out = subtree_logic[index+1].split('=')[0].replace('~', '')
					if bool(pattern.match(out)) == False:
						new_instance = subtree[next_instance_index]
						next_instance_index = next_instance_index + 1	
					elif bool(pattern.match(out)) == True and bool(pattern.match(next_out)) == True and out.split('r')[0] == next_out.split('r')[0]:
						sub_instance_no = out.split('r')[1]
						new_instance = subtree[next_instance_index] + 'r' + sub_instance_no
					else:
						sub_instance_no = out.split('r')[1]
						new_instance = subtree[next_instance_index] + 'r' + sub_instance_no
						next_instance_index = next_instance_index + 1
				optimal_subtree.append(new_instance)

				# update cell type
				optimal_cell_typ.append(cell_typ[matched_index[0]])

	# if there is no match, do nothing
	if match == False:

		optimal_subtree = []
		optimal_cell_typ = []
		optimal_nodes = []

		optimal_subtree = subtree

		optimal_cell_typ.append(root_typ)
		optimal_nodes.append(root_nodes)
		for instance_index, instance in enumerate(subtree[1:]):
			optimal_cell_typ.append(leave_typ[instance_index])
			optimal_nodes.append(leave_nodes[instance_index])

	# original subtree
	org_cell_typ = []
	org_nodes = []

	org_subtree = subtree

	org_cell_typ.append(root_typ)
	org_nodes.append(root_nodes)
	for instance_index, instance in enumerate(subtree[1:]):
		org_cell_typ.append(leave_typ[instance_index])
		org_nodes.append(leave_nodes[instance_index])

	return optimal_subtree, optimal_cell_typ, optimal_nodes, org_subtree, org_cell_typ, org_nodes

# *******************************************************************************


# *******************************************************************************
# compute tree stress
def tree_stress_computation(tree, timing, cp_index, wire_in, port_in, sp_order, node_order, path_stress):

	# capture critical paths/nodes/ports/stress
	all_critical_paths = [timing.instance[i] for i in cp_index]
	all_critical_nodes = [wire_in[i] for i in cp_index]
	all_critical_ports = [port_in[i] for i in cp_index]
	all_critical_stress = [path_stress[i] for i in cp_index]
	
	# remove redundancy
	critical_paths = []
	critical_nodes = []
	critical_ports = []
	critical_stress = []
	for i, path in enumerate(all_critical_paths):
		if path not in critical_paths:
			critical_paths.append(path)
			critical_nodes.append(all_critical_nodes[i])
			critical_ports.append(all_critical_ports[i])
			critical_stress.append(all_critical_stress[i])	

	# find critical paths/nodes/ports/signal probabilities/stress for the tree	
	tree_paths = []
	tree_nodes = []
	tree_ports = []
	tree_sps = []
	tree_stress = []

	root_instance = tree[0]
	cnt = 0

	for i, path in enumerate(critical_paths):

		if root_instance in path:

			root_index = path.index(root_instance)
			tree_paths.append([])
			tree_nodes.append([])
			tree_ports.append([])
			tree_sps.append([])
			tree_stress.append([])

			tree_paths[cnt].append(root_instance)
			tree_nodes[cnt].append(critical_nodes[i][root_index-1])
			tree_ports[cnt].append(critical_ports[i][root_index-1])
			tree_stress[cnt].append(critical_stress[i][root_index-1][0])

			node_sp = sp_order[node_order.index(critical_nodes[i][root_index-1])]
			tree_sps[cnt].append(node_sp)
			
			index = root_index - 1
			while path[index] in tree:
				tree_paths[cnt].append(path[index])
				tree_nodes[cnt].append(critical_nodes[i][index-1])
				tree_ports[cnt].append(critical_ports[i][index-1])
				tree_stress[cnt].append(critical_stress[i][index-1][0])

				node_sp = sp_order[node_order.index(critical_nodes[i][index-1])]
				tree_sps[cnt].append(node_sp)

				index = index - 1
			cnt = cnt + 1

	return tree_paths, tree_nodes, tree_ports, tree_sps, tree_stress

# *******************************************************************************


# *******************************************************************************
# recompute stress and give final subtree final structure
def tree_stress_recomputation(tree_nodes, tree_stress, optimal_subtree, optimal_cell_typ, optimal_nodes, org_subtree, sp_order, node_order, cell_typ_up, stacking_input_up, expression_up, sp_typ, sp_out, sp_expression, new_out_nodes, new_out_sps):

	org_avg_stress0 = []
	org_avg_stress1 = []
	final_structure = ''
	flatten_tree_nodes = [item for sublist in tree_nodes for item in sublist]

	out_nodes = []
	out_sps = []
	out_nodes_irr = []
	out_sps_irr = []
	
	# work out original path average stress
	for each in tree_stress:
		
		x = range(0, len(each))

		# transition 0
		index_trans_0 = filter(lambda x : x % 2 == 0, x)
		stress_trans_0 = [each[i] for i in index_trans_0]
		if len(stress_trans_0) > 0:
			avg_trans_0 = sum(stress_trans_0) / len(stress_trans_0)
		else:
			avg_trans_0 = 0
		org_avg_stress0.append(avg_trans_0)	

		# transition 1
		index_trans_1 = filter(lambda x : x % 2 == 1, x)
		stress_trans_1 = [each[i] for i in index_trans_1]
		if len(stress_trans_1) > 0:
			avg_trans_1 = sum(stress_trans_1) / len(stress_trans_1)
		else:
			avg_trans_1 = 0
		org_avg_stress1.append(avg_trans_1)

	# check if subtree structure changed
	if optimal_subtree == org_subtree:
		final_structure = 'org'

	else:
		in_ct_nodes = []

		# work out driving instances' stress and output sp
		for i, nodes in enumerate(optimal_nodes[1:]):

			i = i + 1
			
			in_nodes = nodes[:-1]
			
			# find input node sps
			in_sps = []
			for in_node in in_nodes:
				if in_node in node_order:
					in_sps.append(sp_order[node_order.index(in_node)])
				else:
					in_sps.append(new_out_sps[new_out_nodes.index(in_node)])

			# check if the gate has critical input nodes
			common_nodes = list(set(flatten_tree_nodes).intersection(in_nodes))
			if len(common_nodes) > 0:

				in_ct_nodes.append(common_nodes)

				# work out critical input node stress duty cycles
				for ct_node in common_nodes:
					inport = string.uppercase[nodes.index(ct_node)]
					
					# check if there is stacking effect
					# pull-up network
					if (optimal_cell_typ[i], inport) in zip(cell_typ_up, stacking_input_up):
				
						# find stacking expression
						expression_index = zip(cell_typ_up, stacking_input_up).index((optimal_cell_typ[i], inport))
						expression = expression_up[expression_index]		

						# find all input pins related to stacking expression
						all_input = [letter for letter in expression if str.isalpha(letter) == True]
						
						# find all input sps related to stacking expression
						stacking_up_sp = [in_sps[string.uppercase.index(pin)] for pin in all_input]

						# work out stress duty cycle
						stress = signal_probability.spout(expression, all_input, stacking_up_sp)

					else:
						stress = 1 - in_sps[in_nodes.index(ct_node)]
						
					# replace old stress with new one
					for ct_nodes_index, ct_nodes in enumerate(tree_nodes):
						if ct_node in ct_nodes:
							tree_stress[ct_nodes_index][ct_nodes.index(ct_node)] = stress

				# work out output node sp
				out_node = nodes[-1]
				_, expression = signal_probability.find_cell(optimal_cell_typ[i], sp_typ, sp_out, sp_expression)
				in_ports = [string.uppercase[index] for index, _ in enumerate(in_nodes)]
				out_sp = signal_probability.spout(expression[0], in_ports, in_sps)

				out_nodes.append(out_node)
				out_sps.append(out_sp)

			else:
				# work out output node sp
				out_node = nodes[-1]
				_, expression = signal_probability.find_cell(optimal_cell_typ[i], sp_typ, sp_out, sp_expression)
				in_ports = [string.uppercase[index] for index, _ in enumerate(in_nodes)]
				out_sp = signal_probability.spout(expression[0], in_ports, in_sps)

				out_nodes_irr.append(out_node)
				out_sps_irr.append(out_sp)

		# work out driven instance stress
		root_nodes = optimal_nodes[0]
		
		# find input node sps
		in_sps = []
		for in_node in root_nodes[:-1]:
			if in_node in node_order:
				sp = sp_order[node_order.index(in_node)]
			elif in_node in out_nodes:
				sp = out_sps[out_nodes.index(in_node)]
			elif in_node in out_nodes_irr:
				sp = out_sps_irr[out_nodes_irr.index(in_node)]
			in_sps.append(sp)

		for out_node_i, out_node in enumerate(out_nodes):
			inport = string.uppercase[root_nodes[:-1].index(out_node)]

			# check if there is stacking effect
			# pull-up network
			if (optimal_cell_typ[0], inport) in zip(cell_typ_up, stacking_input_up):
				
				# find stacking expression
				expression_index = zip(cell_typ_up, stacking_input_up).index((optimal_cell_typ[0], inport))
				expression = expression_up[expression_index]

				# find all input pins related to stacking expression
				all_input = [letter for letter in expression if str.isalpha(letter) == True]
	
				# find all input sps related to stacking expression
				stacking_up_sp = [in_sps[string.uppercase.index(pin)] for pin in all_input]

				# work out stress duty cycle
				stress = signal_probability.spout(expression, all_input, stacking_up_sp)
			
			else:
				stress = 1 - float(in_sps[root_nodes[:-1].index(out_node)])

			# replace old stress with new one
			for ct_nodes_index, ct_nodes in enumerate(tree_nodes):
				if out_node.split('r')[0] in ct_nodes and set(in_ct_nodes[out_node_i]).issubset(set(ct_nodes)):
					old_out_node_index = ct_nodes.index(out_node.split('r')[0])
					tree_nodes[ct_nodes_index][old_out_node_index] = out_node
					tree_stress[ct_nodes_index][old_out_node_index] = stress

		new_avg_stress0 = []
		new_avg_stress1 = []

		# work out original path average stress
		for each in tree_stress:
		
			x = range(0, len(each))

			# transition 0
			index_trans_0 = filter(lambda x : x % 2 == 0, x)
			stress_trans_0 = [each[i] for i in index_trans_0]
			if len(stress_trans_0) > 0:
				avg_trans_0 = sum(stress_trans_0) / len(stress_trans_0)
			else:
				avg_trans_0 = 0
			new_avg_stress0.append(avg_trans_0)	

			# transition 1
			index_trans_1 = filter(lambda x : x % 2 == 1, x)
			stress_trans_1 = [each[i] for i in index_trans_1]
			if len(stress_trans_1) > 0:
				avg_trans_1 = sum(stress_trans_1) / len(stress_trans_1)
			else:
				avg_trans_1 = 0
			new_avg_stress1.append(avg_trans_1)

		# check which structure need to be used
		if max(new_avg_stress0) == max(org_avg_stress0) and max(new_avg_stress1) == max(org_avg_stress1):
			final_structure = 'org'
		elif max(new_avg_stress0) <= max(org_avg_stress0) and max(new_avg_stress1) <= max(org_avg_stress1):
			final_structure = 'new'
		else:
			final_structure = 'org'


	return final_structure, out_nodes, out_sps, out_nodes_irr, out_sps_irr

# *******************************************************************************


# *******************************************************************************
# update tree netlist
def tree_netlist_update(tree_netlist, subtree_line_index, final_subtree, final_cell_typ, final_nodes, lib, tree_netlist_file, critical_instances, netlist):	

	# delete old cell definition lines
	for index in sorted(subtree_line_index, reverse=True):
		tree_netlist.remove(tree_netlist[index])

	# produce new cell definition lines and update netlist
	for index, instance in enumerate(final_subtree):

		# cell type
		typ = final_cell_typ[index]
		
		# cell size
		if instance in critical_instances or 'r' in instance:
			size_index = lib.cell_typ.index(typ)
			all_size = [int(each) for each in lib.cell_size[size_index]]
			size = min(all_size)
		else:
			size_index = netlist.instance_backup.index(instance)
			size = int(netlist.cell_size[size_index])

		nodes = final_nodes[index]

		# cell input
		all_input_def = []
		for i_node, node in enumerate(nodes[:-1]):
			alphabet = string.uppercase[i_node]
			definition = '.%s(%s),' %(alphabet, node)
			all_input_def.append(definition)
		input_def = ' '.join(all_input_def)
		
		# cell output
		output_def = ' .Z(%s)' %nodes[-1]
		
		# entire definition
		cell_def = '  %s%d %s (%s%s);\n' %(typ, size, instance, input_def, output_def)
		
		tree_netlist.insert(min(subtree_line_index), cell_def)

	content = open(tree_netlist_file, 'w')
	for item in tree_netlist:
		content.write(item)
	
# *******************************************************************************


# *******************************************************************************
# restore the tree
def tree_restore(deg_tree_netlist, delay_tree_netlist, tree_index):
	
	# read tree netlists
	deg_tree_content = general.read_txt(deg_tree_netlist)
	delay_tree_content = general.read_txt(delay_tree_netlist)

	tree_name = 'tree' + str(tree_index)
		
	# find tree definition line in delay tree netlist
	tree_line_index = 0
	while len(delay_tree_content[tree_line_index].split()) < 2 or delay_tree_content[tree_line_index].split()[1] != tree_name:
		tree_line_index = tree_line_index + 1

	# find endmodule line in delay tree netlist
	end_line_index = tree_line_index + 1
	while 'endmodule' not in delay_tree_content[end_line_index]:
		end_line_index = end_line_index + 1

	# find tree definition line in degradation tree netlist
	tree_line_index1 = 0
	while len(deg_tree_content[tree_line_index1].split()) < 2 or deg_tree_content[tree_line_index1].split()[1] != tree_name:
		tree_line_index1 = tree_line_index1 + 1

	# remove the tree from the netlist
	while 'endmodule' not in deg_tree_content[tree_line_index1]:
		deg_tree_content.pop(tree_line_index1)
	deg_tree_content.pop(tree_line_index1)

	# insert tree definition lines from delay netlist ot degradation one
	for i in range(end_line_index,tree_line_index-1,-1):
		deg_tree_content.insert(tree_line_index1, delay_tree_content[i])

	# update mininal degradation tree netlist
	deg_file = open(deg_tree_netlist, 'w')
	for item in deg_tree_content:
		deg_file.write(item)

# *******************************************************************************


# *******************************************************************************
# update tree synthesis scripts
def tree_tcl_update(tree_tcl, area_cons, first_time, delay_line_index, delay_cons, tree_index, non_critical_instances):

	new_delay_line_index = 0
	tree_name = 'tree' + str(tree_index)
	tree_tcl_content = general.read_txt(tree_tcl)

	if first_time == True:
		new_tree_tcl_content = []
		for index, line in enumerate(tree_tcl_content):
		
				# change tree name
				if 'elaborate' in line:
					old_tree_name = line.split()[1]
					new_line = line.replace(old_tree_name, tree_name)

				# change area constraint
				elif 'set_max_area' in line:
					new_line = 'set_max_area %f\n' %area_cons

				# change untouched instance
				elif 'set_dont_touch' in line:
					instance_list = ' '.join(non_critical_instances)
					new_line = 'set_dont_touch {%s}\n' %instance_list
		
				# change delay constraint
				elif 'set_max_delay' in line:
					new_line = 'set_max_delay %f -from [all_inputs]\n' %delay_cons
					new_delay_line_index = index
		
				else:
					new_line = line

				new_tree_tcl_content.append(new_line)

	else:
		new_tree_tcl_content = tree_tcl_content
		new_tree_tcl_content[delay_line_index] = 'set_max_delay %f -from [all_inputs]\n' %delay_cons

	tcl_file = open(tree_tcl, 'w')
	for item in new_tree_tcl_content:
		tcl_file.write(item)

	return new_delay_line_index

# *******************************************************************************
	

# *******************************************************************************
# gate count
def gate_count(tree_netlist, tree_index):

	# read tree netlists
	tree_content = general.read_txt(tree_netlist)

	tree_name = 'tree' + str(tree_index)
		
	# find tree definition line in tree netlist
	tree_line_index = 0
	while len(tree_content[tree_line_index].split()) < 2 or tree_content[tree_line_index].split()[1] != tree_name:
		tree_line_index = tree_line_index + 1

	# find endmodule line in tree netlist
	end_line_index = tree_line_index + 1
	while 'endmodule' not in tree_content[end_line_index]:
		end_line_index = end_line_index + 1

	# count gate
	pattern1 = re.compile('(^U[0-9]+)')
	pattern2 = re.compile('(^U[0-9]+r[0-9+])')
	number_of_gate = 0
	for line in tree_content[tree_line_index:end_line_index+1]:
		if len(line.split()) > 1:
			instance = line.split()[1]	
			if (bool(pattern1.match(instance)) == True) or (bool(pattern2.match(instance)) == True):
				number_of_gate = number_of_gate + 1

	return number_of_gate

# *******************************************************************************


# *******************************************************************************
# replace the tree in original netlist with optimized one
def org_netlist_replace(org_netlist, backup_netlist, min_deg_tree_netlist, tree_index, tree):

	# read original netlist
	netlist_content = general.read_txt(org_netlist)

	# write backup
	shutil.copy(org_netlist, backup_netlist)

	# read minimal degradation tree netlist
	tree_content = general.read_txt(min_deg_tree_netlist)

	# delete line containing tree instances
	new_netlist_content = []
	semicolon_removal = False
	for line_index, line in enumerate(netlist_content):

		if len(line.split()) > 1:
			instance = line.split()[1]
			if instance not in tree and semicolon_removal == False:
				new_netlist_content.append(line)
			else:
				if ';' not in line:
					semicolon_removal = True
				else:
					semicolon_removal = False
				
		elif semicolon_removal == False:
			new_netlist_content.append(line)

		elif semicolon_removal == True and ';' in line:
			semicolon_removal = False

	# find tree definition line in tree netlist
	tree_line_index = 0	
	tree_name = 'tree' + str(tree_index)
	while len(tree_content[tree_line_index].split()) < 2 or tree_content[tree_line_index].split()[1] != tree_name:
		tree_line_index = tree_line_index + 1

	# add tree definition into the netlist
	endmodule_index = [index for index, line in enumerate(new_netlist_content) if 'endmodule' in line][0]
	if ';' in tree_content[tree_line_index]:
		tree_def = tree_content[tree_line_index].replace('module', '')
	else:
		tree_def = ''
		line_index_cnt = tree_line_index
		while ';' not in tree_content[line_index_cnt]:
			tree_def = tree_def + tree_content[line_index_cnt]
			line_index_cnt = line_index_cnt + 1
		tree_def = tree_def + tree_content[line_index_cnt]
		tree_def = tree_def.replace('module', '')
	tree_def_split = tree_def.split()
	tree_def_split.insert(1, 'T'+str(tree_index))
	tree_def = '\t'.join(tree_def_split)
	new_netlist_content.insert(endmodule_index, tree_def+'\n')
	
	netlist_file = open(org_netlist, 'w')
	for line in new_netlist_content:
		netlist_file.write(line)

# *******************************************************************************


# *******************************************************************************
# update backward pass scripts
def backward_tcl_update(backward_tcl, netlist_file_list, tree_instance, root_node):
	
	backward_tcl_content = general.read_txt(backward_tcl)
	new_backward_content = []

	for line in backward_tcl_content:
		
		if 'analyze' in line:
			all_netlist_files = '\t'.join(netlist_file_list)
			new_line = 'analyze -format verilog {%s}\n' %all_netlist_files

		elif 'rise_from' in line:
			new_line = 'report_timing -max_path 1 -nworst 1 -significant_digits 5 -rise_from {%s/%s} > backward_rise.rpt\n' %(tree_instance, root_node)

		elif 'fall_from' in line:
			new_line = 'report_timing -max_path 1 -nworst 1 -significant_digits 5 -fall_from {%s/%s} > backward_fall.rpt\n' %(tree_instance, root_node)

		else:
			new_line = line

		new_backward_content.append(new_line)

	tcl_file = open(backward_tcl, 'w')
	for line in new_backward_content:
		tcl_file.write(line)

# *******************************************************************************


# *******************************************************************************
# check arrival time/required arrival time for tree root 
def arrival_time_checking(timing_cons, tree_instance, root_node):

	backward_rise_content = general.read_txt('backward_rise.rpt')
	backward_fall_content = general.read_txt('backward_fall.rpt')
	backward_content = backward_rise_content + backward_fall_content

	node_name = tree_instance + '/' + root_node

	all_root_arrival = []
	all_path_arrival = []

	for line in backward_content:
		
		# find root node arrival time 
		if '<-' in line and node_name in line:
			root_arrival = line.split()[4]
			all_root_arrival.append(root_arrival)

		# find path arrival time
		elif 'data arrival time' in line:
			path_arrival = line.split()[3]
			all_path_arrival.append(path_arrival)
	
	# path arrival time after root node
	all_remain_arrival = [float(all_path_arrival[i]) - float(all_root_arrival[i]) for i in range(0, len(all_path_arrival))]

	# required arrival time
	all_req_arrival = [(timing_cons - time) for time in all_remain_arrival]

	return all_root_arrival, all_req_arrival

# *******************************************************************************
